from typing import Optional

from genai._utils.service import (
    BaseService,
    BaseServiceConfig,
    BaseServiceServices,
    get_service_action_metadata,
    set_service_action_metadata,
)
from genai.schema import (
    SystemPromptCreateEndpoint,
    SystemPromptCreateResponse,
    SystemPromptIdDeleteEndpoint,
    SystemPromptIdRetrieveEndpoint,
    SystemPromptIdRetrieveResponse,
    SystemPromptIdUpdateEndpoint,
    SystemPromptIdUpdateResponse,
    SystemPromptRetrieveEndpoint,
    SystemPromptRetrieveResponse,
)
from genai.schema._api import (
    _SystemPromptCreateParametersQuery,
    _SystemPromptCreateRequest,
    _SystemPromptIdDeleteParametersQuery,
    _SystemPromptIdRetrieveParametersQuery,
    _SystemPromptIdUpdateParametersQuery,
    _SystemPromptIdUpdateRequest,
    _SystemPromptRetrieveParametersQuery,
)

__all__ = ["SystemPromptService"]


class SystemPromptService(BaseService[BaseServiceConfig, BaseServiceServices]):
    @set_service_action_metadata(endpoint=SystemPromptCreateEndpoint)
    def create(self, *, name: str, content: str) -> SystemPromptCreateResponse:
        """
        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
            ValidationError: In case of provided parameters are invalid.
        """
        request_body = _SystemPromptCreateRequest(name=name, content=content).model_dump()

        self._log_method_execution("System Prompt Create", **request_body)

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.create)
            response = client.post(
                url=self._get_endpoint(metadata.endpoint),
                params=_SystemPromptCreateParametersQuery().model_dump(),
                json=request_body,
            )
            return SystemPromptCreateResponse(**response.json())

    @set_service_action_metadata(endpoint=SystemPromptIdRetrieveEndpoint)
    def retrieve(
        self,
        id: int,
    ) -> SystemPromptIdRetrieveResponse:
        """
        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
            ValidationError: In case of provided parameters are invalid.
        """
        self._log_method_execution("System Prompt Retrieve", id=id)

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.retrieve)
            response = client.get(
                url=self._get_endpoint(metadata.endpoint, id=id),
                params=_SystemPromptIdRetrieveParametersQuery().model_dump(),
            )
            return SystemPromptIdRetrieveResponse(**response.json())

    @set_service_action_metadata(endpoint=SystemPromptIdUpdateEndpoint)
    def update(self, id: int, *, name: str, content: str) -> SystemPromptIdUpdateResponse:
        """
        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
            ValidationError: In case of provided parameters are invalid.
        """
        request_body = _SystemPromptIdUpdateRequest(name=name, content=content).model_dump()

        self._log_method_execution("System Prompt Update", **request_body)

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.update)
            response = client.post(
                url=self._get_endpoint(metadata.endpoint, id=id),
                params=_SystemPromptIdUpdateParametersQuery().model_dump(),
                json=request_body,
            )
            return SystemPromptIdUpdateResponse(**response.json())

    @set_service_action_metadata(endpoint=SystemPromptRetrieveEndpoint)
    def list(
        self,
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> SystemPromptRetrieveResponse:
        """
        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
            ValidationError: In case of provided parameters are invalid.
        """
        request_parameters = _SystemPromptRetrieveParametersQuery(
            limit=limit,
            offset=offset,
        ).model_dump()
        self._log_method_execution("System Prompt List", **request_parameters)

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.list)
            response = client.get(url=self._get_endpoint(metadata.endpoint), params=request_parameters)
            return SystemPromptRetrieveResponse(**response.json())

    @set_service_action_metadata(endpoint=SystemPromptIdDeleteEndpoint)
    def delete(
        self,
        id: int,
    ) -> None:
        """
        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
            ValidationError: In case of provided parameters are invalid.
        """
        self._log_method_execution("System Prompt Delete", id=id)

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.delete)
            client.delete(
                url=self._get_endpoint(metadata.endpoint, id=id),
                params=_SystemPromptIdDeleteParametersQuery().model_dump(),
            )



import asyncio
from typing import Callable, Optional

from genai._utils.limiters.base_limiter import BaseLimiter


class LoopBoundLimiter(BaseLimiter):
    """
    A class that limits the frequency of executing a code block using a provided limiter.
    The limiter must be created inside an event loop.
    """

    _limiter: BaseLimiter
    _loop: Optional[asyncio.AbstractEventLoop] = None

    def __init__(self, factory: Callable[[], BaseLimiter]):
        self._factory = factory

    async def acquire(self):
        return await self._get_limiter().acquire()

    def release(self):
        return self._get_limiter().release()

    async def report_error(self):
        return await self._limiter.report_error()

    async def report_success(self):
        return await self._limiter.report_success()

    def _get_limiter(self) -> BaseLimiter:
        loop = asyncio.get_running_loop()
        if loop is not self._loop:
            self._limiter = self._factory()
            self._loop = loop
        return self._limiter



from typing import Optional

from aiolimiter import AsyncLimiter
from httpx import Request, Response

__all__ = ["AsyncRateLimiter", "AsyncRateLimitTransport"]

from genai._utils.http_client.retry_transport import AsyncRetryTransport


class AsyncRateLimiter(AsyncLimiter):
    """Modified version of AsyncLimiter with no-bursting approach and customizable limits"""

    def __init__(self, max_rate: float, time_period: float) -> None:
        super().__init__(max_rate=1, time_period=time_period / max_rate)
        self._source_max_rate = max_rate
        self._source_time_period = time_period

    def update_limit(self, *, max_rate: Optional[float] = None, time_period: Optional[float] = None) -> None:
        """Updates the maximum rate and time period for the source.

        Args:
            max_rate: The maximum rate for the source.
            time_period: The time period for the source.
        """
        if max_rate is None:
            max_rate = self._source_max_rate
        else:
            self._source_max_rate = max_rate

        if time_period is None:
            time_period = self._source_time_period
        else:
            self._source_time_period = time_period

        self.max_rate = 1
        self.time_period = time_period / max_rate
        self._rate_per_sec = self.max_rate / self.time_period


class AsyncRateLimitTransport(AsyncRetryTransport):
    """
    Transport class that enforces rate limiting for asynchronous requests.

    Args:
        *args: Variable length argument list passed to `AsyncRetryTransport`.
        max_rate: The default maximum rate allowed for requests (in requests per time_period).
        time_period: The default time period for rate limiting (in seconds).
        disable_rate_limit_no_header: disable rate limiting if the response header has no rate limit information.
        **kwargs: Keyword arguments passed to AsyncRetryTransport.
    """

    def __init__(
        self,
        *args,
        max_rate: float,
        time_period: float = 1.0,
        disable_rate_limit_no_header: bool = False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._active: bool = True
        self._disable_rate_limit_no_header = disable_rate_limit_no_header
        self._limiter = AsyncRateLimiter(max_rate=max_rate, time_period=time_period)

    def _update_rate_limit(self, response: Response):
        max_rate = response.headers.get("x-ratelimit-limit")

        if max_rate is not None:
            self._active = True
            self._limiter.update_limit(
                max_rate=max(int(max_rate), 1),
            )
        elif self._disable_rate_limit_no_header:
            self._active = False

    async def handle_async_request(
        self,
        request: Request,
    ) -> Response:
        if not self._active:
            return await super().handle_async_request(request)

        async with self._limiter:
            response = await super().handle_async_request(request)
            self._update_rate_limit(response)
            return response



import re

__all__ = ["to_langchain_template", "from_langchain_template"]


def to_langchain_template(template: str) -> str:
    """Convert mustache template variables to langchain template variables"""
    return re.sub(r"{{([^}]+)}}", r"{\1}", template, count=0, flags=re.MULTILINE)


def from_langchain_template(template: str) -> str:
    """Convert langchain template variables to mustache template variables"""
    return re.sub(r"{([^}]+)}", r"{{\1}}", template, count=0, flags=re.MULTILINE)



from genai._utils.service import (
    BaseService,
    BaseServiceConfig,
    BaseServiceServices,
    get_service_action_metadata,
    set_service_action_metadata,
)
from genai.schema import TextGenerationLimitRetrieveEndpoint, TextGenerationLimitRetrieveResponse
from genai.schema._api import _TextGenerationLimitRetrieveParametersQuery

__all__ = ["LimitService"]


class LimitService(BaseService[BaseServiceConfig, BaseServiceServices]):
    @set_service_action_metadata(endpoint=TextGenerationLimitRetrieveEndpoint)
    def retrieve(self) -> TextGenerationLimitRetrieveResponse:
        """
        Retrieves the current text generation limit from the server.

        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
        """
        self._log_method_execution("Text Generation Limit Retrieve")

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.retrieve)
            response = client.get(
                url=self._get_endpoint(metadata.endpoint),
                params=_TextGenerationLimitRetrieveParametersQuery().model_dump(),
            )
            return TextGenerationLimitRetrieveResponse(**response.json())

    @set_service_action_metadata(endpoint=TextGenerationLimitRetrieveEndpoint)
    async def aretrieve(self) -> TextGenerationLimitRetrieveResponse:
        """
        Retrieves the current text generation limit from the server.

        Raises:
            ApiResponseException: If target feedback/generation does not exist or cannot be updated.
            ApiNetworkException: In case of unhandled network error.
        """
        self._log_method_execution("Text Generation Limit ARetrieve")

        async with self._get_async_http_client() as client:
            metadata = get_service_action_metadata(self.aretrieve)
            response = await client.get(
                url=self._get_endpoint(metadata.endpoint),
                params=_TextGenerationLimitRetrieveParametersQuery().model_dump(),
            )
            return TextGenerationLimitRetrieveResponse(**response.json())



import logging
from typing import Optional, Union

from httpx import HTTPError, Response
from pydantic import ValidationError as _ValidationError
from typing_extensions import TypeAlias

from genai._utils.responses import is_api_error_response, to_api_error
from genai.schema import BaseErrorResponse

logger = logging.getLogger(__name__)

__all__ = ["BaseApiException", "ApiResponseException", "ApiNetworkException", "ValidationError"]


ValidationError: TypeAlias = _ValidationError


class BaseApiException(Exception):
    """
    Exception class for API related errors.
    """

    message: str


class ApiResponseException(BaseApiException):
    """
    Exception class for API with valid response body.
    """

    response: BaseErrorResponse

    def __init__(
        self,
        response: Union[BaseErrorResponse, dict],
        message: Optional[str] = None,
        *args,
    ) -> None:
        if is_api_error_response(response):
            self.response = response
        elif isinstance(response, dict):
            self.response = to_api_error(response)
        else:
            raise TypeError(f"Expected either Response or Api Error Response, but {type(response)} received.")

        self.message = f"{message or 'Server Error'}\n{self.response.model_dump_json(indent=2)}"
        super().__init__(self.message, *args)

    @classmethod
    def from_http_response(cls, response: Response, message: Optional[str] = None):
        if response.is_success:
            raise ValueError("Cannot convert succeed HTTP response to error response.")

        response_body = to_api_error(response.json())
        return cls(message=message, response=response_body)

    def __reduce__(self):
        return self.__class__, (self.response.model_dump(), self.message)


class ApiNetworkException(BaseApiException):
    """
    Exception raised when there is a network-related error during API communication ('httpx' related error).

    Attributes:
        message (str): Explanation of the error.
    """

    __cause__: Optional[HTTPError] = None

    def __init__(self, message: Optional[str] = None, *args) -> None:
        self.message = message or "Network Exception has occurred. Try again later."
        super().__init__(self.message, *args)



import warnings

_CACHED_WARNINGS: set[str] = set()


def _log_deprecation_warning(name: str, module_name: str):
    cache_warning_name = f"{module_name}.{name}"

    if cache_warning_name in _CACHED_WARNINGS:
        return

    msg = f"Deprecated import of {name} from module {module_name}. Please use `from genai.schema import {name}`."
    with warnings.catch_warnings():
        warnings.simplefilter("always", DeprecationWarning)
        warnings.warn(msg, category=DeprecationWarning, stacklevel=4)  # the original import is 4 levels higher
    _CACHED_WARNINGS.add(cache_warning_name)


def _deprecated_schema_import(name: str, module_name: str):
    """
    Support for deprecated import style "from genai.text.generation import" -> should be "from genai.schema import"

    TODO(#297): to be removed in next major version
    """
    import genai.schema

    if name in dir(genai.schema) and not name.startswith("_"):
        _log_deprecation_warning(name, module_name)
        return getattr(genai.schema, name)

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



import functools
import hashlib
import json
from enum import Enum
from typing import Any, Callable, Generator, Mapping, Optional, Type, TypeVar, Union

from pydantic import BaseModel

from genai._types import EnumLike

TModel = TypeVar("TModel", bound=BaseModel)


def to_model_optional(
    input_params: Union[Any, list[Any]], Model: Type[TModel], *, copy: bool = True
) -> Optional[TModel]:
    return to_model_instance(input_params, Model, copy=copy) if input_params else None


def to_model_instance(input_params: Union[Any, list[Any]], Model: Type[TModel], *, copy: bool = True) -> TModel:
    """
    Converts input parameters to an instance of the specified Pydantic model.

    Args:
        input_params (Union[Any, list[Any]]): The input parameters to convert.
        Model (Type[TModel]): The type of model to convert to.
        copy (bool, optional): Determines whether to create a copy of the model instance. Defaults to True.

    Returns:
        TModel: An instance of the specified model.

    Raises:
        ValueError: If the 'params' property is not an instance of the specified Model class.

    """
    if not input_params:
        return Model()

    if isinstance(input_params, list):
        result = merge_objects(
            *[
                (
                    obj
                    if isinstance(obj, dict)
                    else to_model_instance(obj, Model).model_dump(exclude_none=True, exclude_defaults=True)
                )
                for obj in input_params
                if obj
            ]
        )
        return Model(**result)

    params = input_params
    if isinstance(params, Model):
        if not copy:
            return params

        return params.model_copy()

    if isinstance(params, BaseModel):
        return Model(**params.model_dump())

    if isinstance(params, dict):
        return Model(**params)

    raise ValueError(f"The 'params' property should be an instance of {Model.__name__} class. Current value: {params}")


def to_model_instance_if_defined(
    input_params: Union[Any, list[Any]], Model: Type[TModel], *, copy: bool = True
) -> Union[TModel, None]:
    return to_model_instance(input_params, Model, copy=copy) if input_params is not None else None


TInput = TypeVar("TInput")


def batch(inputs: list[TInput], *, chunk_size: int) -> list[list[TInput]]:
    if chunk_size <= 0:
        raise ValueError("'chunk_size' must be > 0")

    return [inputs[i : i + chunk_size] for i in range(0, len(inputs), chunk_size)]


def batch_by_size_constraint(
    inputs: list[str], *, max_chunk_size: Optional[int] = None, max_size_bytes: Optional[int] = None
) -> Generator[list[str], None, None]:
    """
    Args:
        inputs: A list of strings representing the inputs to be batched.
        max_chunk_size: An optional integer representing the maximum number of inputs in each batch.
        max_size_bytes: An optional integer representing the maximum total size in bytes for each batch.

    Returns:
        A generator that yields batches of inputs, where each batch is a list of strings.

    Raises:
        ValueError: If both max_chunk_size and max_size_bytes are None.
        ValueError: If max_chunk_size is less than 1.
        ValueError: If max_size_bytes is less than 1.
        ValueError: If a single input is larger than the maximal payload size.

    """
    if max_size_bytes is None and max_chunk_size is None:
        yield inputs.copy()
        return

    if max_size_bytes is None:
        yield from batch(inputs, chunk_size=max_chunk_size or len(inputs))
        return

    if max_chunk_size is None:
        max_chunk_size = len(inputs)

    if max_chunk_size < 1:
        raise ValueError("'chunk_size' must be >= 1")

    if max_size_bytes is None or max_size_bytes < 1:
        raise ValueError("'max_size_bytes' must be >= 1")

    chunk: list[str] = []
    chunk_size = 0

    for input in inputs:
        size = len(input.encode("utf-8"))
        if chunk_size + size > max_size_bytes or len(chunk) == max_chunk_size:
            if not chunk:
                raise ValueError(
                    f"Single input is larger than maximal payload size. Input size: {size}B. Allowed size: {max_size_bytes}B"  # noqa
                )

            yield chunk
            chunk = []
            chunk_size = 0

        chunk.append(input)
        chunk_size += size

    if chunk:
        yield chunk


_KT = TypeVar("_KT")
_VT = TypeVar("_VT")


def merge_objects(*objs: Optional[Mapping[_KT, _VT]]) -> dict[_KT, _VT]:
    result: dict[_KT, _VT] = {}
    for obj in objs:
        if obj is not None:
            result.update(obj)
    return result


def cast_list(input: Union[TInput, list[TInput]]) -> list[TInput]:
    return input if isinstance(input, list) else [input]


def first_defined(*args: Optional[TInput], default: TInput) -> TInput:
    return next((v for v in args if v is not None), default)


E = TypeVar("E", bound=Enum)


def to_enum(cls: type[E], value: EnumLike[E]) -> E:
    return cls(value)


def to_enum_optional(value: Optional[EnumLike[E]], cls: type[E]) -> Optional[E]:
    return None if value is None else cls(value)


def hash_params(**kwargs):
    dhash = hashlib.md5()
    dhash.update(json.dumps(kwargs, sort_keys=True, default=str).encode())
    return dhash.hexdigest()


T = TypeVar("T", bound=Callable)


def single_execution(fn: T) -> T:
    return functools.cache(fn)  # type: ignore



import threading
from abc import abstractmethod
from contextlib import AbstractAsyncContextManager, AbstractContextManager
from typing import Generic, Optional, TypeVar

T = TypeVar("T")


class SharedResource(Generic[T], AbstractContextManager):
    """Class representing a shared resource.

    This class provides a way to manage a shared resource using the context manager protocol.
    It keeps track of the number of references to the resource and initializes it when the first
    reference is made. The resource is released when the last reference is closed.
    """

    def __init__(self):
        self._ref_count = 0
        self._resource: Optional[T] = None
        self._lock = threading.Lock()

    @abstractmethod
    def _enter(self) -> T:
        """
        Create a new resource (called when the first reference is made)

        Returns:
            T: new resource
        """
        raise NotImplementedError

    @abstractmethod
    def _exit(self) -> None:
        """
        Destroys an existing resource (called when there is not a reference to the resource)
        """
        raise NotImplementedError

    def __enter__(self) -> T:
        with self._lock:
            self._ref_count += 1
            if self._ref_count == 1:
                self._resource = self._enter()

            assert self._resource
            return self._resource

    def __exit__(self, exc_type, exc_val, exc_tb):
        with self._lock:
            self._ref_count -= 1
            if self._ref_count == 0:
                self._exit()
                self._resource = None


class AsyncSharedResource(Generic[T], AbstractAsyncContextManager):
    def __init__(self):
        self._ref_count = 0
        self._resource: Optional[T] = None

    @abstractmethod
    async def _enter(self) -> T:
        """
        Create a new resource (called when the first reference is made)

        Returns:
            T: new resource
        """
        raise NotImplementedError

    @abstractmethod
    async def _exit(self) -> T:
        """
        Destroys an existing resource (called when there is not a reference to the resource)
        """
        raise NotImplementedError

    async def __aenter__(self) -> T:
        self._ref_count += 1
        if self._ref_count == 1:
            self._resource = await self._enter()

        assert self._resource
        return self._resource

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        self._ref_count -= 1
        if self._ref_count == 0:
            await self._exit()
            self._resource = None



from asyncio import AbstractEventLoop
from typing import Any, Optional

from httpx import Timeout
from pydantic import BaseModel, ConfigDict, Field, field_validator

from genai._types import ModelLike
from genai._utils.general import hash_params, merge_objects, to_model_instance
from genai._utils.http_client.httpx_client import (
    AsyncHttpxClient,
    HttpxClient,
    ReusableAsyncHttpxClient,
)
from genai._utils.http_client.rate_limit_transport import AsyncRateLimitTransport
from genai._utils.http_client.retry_transport import AsyncRetryTransport, RetryTransport
from genai._utils.shared_loop import shared_event_loop
from genai._version import __version__
from genai.credentials import Credentials

__all__ = ["HttpClientOptions", "HttpTransportOptions", "BaseConfig"]


class HttpClientOptions(BaseModel):
    """Options for httpx client"""

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)

    timeout: Timeout = Timeout(timeout=10 * 60, connect=10)

    @field_validator("timeout", mode="before")
    @classmethod
    def check_timeout(cls, timeout: Any):
        if isinstance(timeout, dict):
            return Timeout(**timeout)
        elif isinstance(timeout, int) or isinstance(timeout, float):
            return Timeout(timeout=timeout, connect=10)

        return timeout


class HttpTransportOptions(BaseModel, extra="allow"):
    """Options for httpx transports"""

    retries: int = 3
    retry_status_codes: Optional[list[int]] = None
    backoff_factor: float = 0.2


class BaseConfig(BaseModel, extra="forbid"):
    """Default configuration for ApiClient"""

    client_options: HttpClientOptions = HttpClientOptions()
    transport_options: HttpTransportOptions = HttpTransportOptions()
    max_payload_size_bytes: int = Field(
        1024**2 // 2,  # 0.5 MiB
        description="Max payload size of body payload which API can process.",
    )


class ApiClient:
    """
    Class which provides methods for making API requests.
    It provides methods for obtaining synchronous and asynchronous HTTP clients.
    """

    Config: type[BaseConfig] = BaseConfig

    def __init__(self, *, credentials: Credentials, config: Optional[ModelLike[BaseConfig]] = None):
        self.config = to_model_instance(config, ApiClient.Config)
        self._credentials = credentials
        self._async_http_clients: dict[str, ReusableAsyncHttpxClient] = {}
        self._latest_event_loop: Optional[AbstractEventLoop] = None

    def get_http_client(self, retry_options: Optional[dict] = None, **kwargs) -> HttpxClient:
        return HttpxClient(
            **self._get_client_options(kwargs),
            transport=RetryTransport(**self._get_transport_options(override=retry_options)),
        )

    def get_async_http_client(
        self,
        rate_limit_options: Optional[dict] = None,
        retry_options: Optional[dict] = None,
        client_options: Optional[dict] = None,
        **kwargs,
    ) -> AsyncHttpxClient:
        """
        Gets cached instance of AsyncHttpxClient (cache key is created from provided parameters).

        Raises:
            ValueError: If both rate_limit_options and retry_options are provided.
            RuntimeError: If the method is not called within an asynchronous environment.
        """
        if rate_limit_options and retry_options:
            raise ValueError("Cannot pass both 'rate_limit_options' and 'retry_options' parameters.")

        loop = shared_event_loop.get_running_loop()
        if loop is None:
            raise RuntimeError("Async HTTP Client must be retrieved within async environment.")

        if self._latest_event_loop is not loop:
            self._async_http_clients.clear()
            self._latest_event_loop = loop

        params_hash = hash_params(
            rate_limit_options=rate_limit_options,
            retry_options=retry_options,
            client_options=client_options,
            **kwargs,
        )

        if params_hash not in self._async_http_clients:
            options = self._get_transport_options(override=rate_limit_options or retry_options)
            transport = AsyncRateLimitTransport(**options) if rate_limit_options else AsyncRetryTransport(**options)

            client = ReusableAsyncHttpxClient(
                **kwargs,
                **self._get_client_options(override=client_options),
                transport=transport,
                on_before_close_callback=lambda: self._async_http_clients.pop(params_hash),
            )
            self._async_http_clients[params_hash] = client

        http_client = self._async_http_clients.get(params_hash)
        assert isinstance(http_client, AsyncHttpxClient)
        assert not http_client.is_closed
        return http_client

    def _get_headers(self, override: Optional[dict]) -> dict:
        headers = {
            **(override or {}),
            "Authorization": f"Bearer {self._credentials.api_key.get_secret_value()}",
            "x-request-origin": f"python-sdk/{__version__}",
            "user-agent": f"python-sdk/{__version__}",
        }

        return headers

    def _get_client_options(self, override: Optional[dict] = None) -> dict:
        final = merge_objects(
            self.config.client_options.model_dump(exclude_none=True),
            override,
            {
                "base_url": self._credentials.api_endpoint,
                "headers": self._get_headers(override=(override or {}).get("headers")),
            },
        )
        return HttpClientOptions.model_validate(final).model_dump()

    def _get_transport_options(self, override: Optional[dict] = None) -> dict:
        return merge_objects(self.config.transport_options.model_dump(exclude_none=True), override)



"""Modules containing functionalities related to text requests (their history)"""

from genai.request.request_service import *


def __getattr__(name: str):
    """Allow additional imports for backward compatibility with old import system "from genai.service_name import"."""
    if name in globals():
        return globals()[name]

    from genai._utils.deprecated_schema_import import _deprecated_schema_import

    return _deprecated_schema_import(name, __name__)



import asyncio
from typing import Optional

from genai._utils.asyncio_future import AsyncioLock
from genai._utils.limiters.base_limiter import BaseLimiter


class LimiterContainer(BaseLimiter):
    """
    A class that represents a container for multiple limiters.
    """

    def __init__(self, *limiters: Optional[BaseLimiter]):
        self._limiters = [limiter for limiter in limiters if limiter]
        self._lock = AsyncioLock()

    async def acquire(self):
        """
        Acquires all the limiters in parallel.

        If there are no limiters available, this method will return True.
        If there is only one limiter available, it will be acquired and the method will return True.
        If there are multiple limiters available, they will be acquired in parallel using asyncio.gather().

        Returns:
            True: If all limiters are acquired successfully.

        Example::
            limiter_group = LimiterGroup()
            await limiter_group.acquire()
        """
        if len(self._limiters) == 0:
            return True

        if len(self._limiters) == 1:
            await self._limiters[0].acquire()
            return True

        async with self._lock:
            await asyncio.gather(*[limiter.acquire() for limiter in self._limiters])
            return True

    def release(self):
        """
        Release all limiters.
        """
        for limiter in self._limiters:
            limiter.release()

    async def report_error(self):
        """
        Report error to nested limiters.

        This method asynchronously reports any errors that occurred during execution.
        """
        await asyncio.gather(*[limiter.report_error() for limiter in self._limiters])

    async def report_success(self):
        """
        Report success to nested limiters.
        """
        await asyncio.gather(*[limiter.report_success() for limiter in self._limiters])



from typing import Optional, Union

from genai._types import EnumLike, EnumLikeOrEnumLikeList, ModelLike
from genai._utils.general import cast_list, to_enum, to_enum_optional, to_model_instance, to_model_optional
from genai._utils.service import (
    BaseService,
    BaseServiceConfig,
    BaseServiceServices,
    get_service_action_metadata,
    set_service_action_metadata,
)
from genai._utils.validators import assert_is_not_empty_string
from genai.schema import (
    BaseMessage,
    ModerationParameters,
    PromptCreateResponse,
    PromptIdRetrieveResponse,
    PromptIdUpdateResponse,
    PromptRetrieveRequestParamsSource,
    PromptRetrieveResponse,
    PromptTemplateData,
    PromptType,
    TextGenerationParameters,
)
from genai.schema._api import (
    _PromptCreateParametersQuery,
    _PromptCreateRequest,
    _PromptIdDeleteParametersQuery,
    _PromptIdRetrieveParametersQuery,
    _PromptIdUpdateParametersQuery,
    _PromptIdUpdateRequest,
    _PromptRetrieveParametersQuery,
)
from genai.schema._endpoints import (
    PromptCreateEndpoint,
    PromptIdDeleteEndpoint,
    PromptIdRetrieveEndpoint,
    PromptIdUpdateEndpoint,
    PromptRetrieveEndpoint,
)

__all__ = ["PromptService"]


class PromptService(BaseService[BaseServiceConfig, BaseServiceServices]):
    @set_service_action_metadata(endpoint=PromptCreateEndpoint)
    def create(
        self,
        *,
        name: str,
        model_id: str,
        prompt_id: Optional[str] = None,
        messages: Optional[list[ModelLike[BaseMessage]]] = None,
        task_id: Optional[str] = None,
        description: Optional[str] = None,
        moderations: Optional[ModelLike[ModerationParameters]] = None,
        data: Optional[ModelLike[PromptTemplateData]] = None,
        type: Optional[EnumLike[PromptType]] = None,
        input: Optional[str] = None,
        output: Optional[str] = None,
        parameters: Optional[ModelLike[TextGenerationParameters]] = None,
    ) -> PromptCreateResponse:
        """
        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
            ValidationError: In case of provided parameters are invalid.
        """
        request_body = _PromptCreateRequest(
            name=name,
            model_id=model_id,
            prompt_id=prompt_id,
            messages=[to_model_instance(msg, BaseMessage) for msg in messages] if messages else None,
            task_id=task_id,
            description=description,
            moderations=to_model_optional(moderations, ModerationParameters),
            data=to_model_optional(data, PromptTemplateData),
            input=input,
            output=output,
            parameters=to_model_optional(parameters, TextGenerationParameters),
            type=to_enum_optional(type, PromptType),
        ).model_dump()

        self._log_method_execution("Prompts Create", **request_body)

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.create)
            response = client.post(
                url=self._get_endpoint(metadata.endpoint),
                params=_PromptCreateParametersQuery().model_dump(),
                json=request_body,
            )
            return PromptCreateResponse(**response.json())

    @set_service_action_metadata(endpoint=PromptIdRetrieveEndpoint)
    def retrieve(
        self,
        id: str,
    ) -> PromptIdRetrieveResponse:
        """
        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
            ValidationError: In case of provided parameters are invalid.
        """
        assert_is_not_empty_string(id)
        self._log_method_execution("Prompts Retrieve", id=id)

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.retrieve)
            response = client.get(
                url=self._get_endpoint(metadata.endpoint, id=id),
                params=_PromptIdRetrieveParametersQuery().model_dump(),
            )
            return PromptIdRetrieveResponse(**response.json())

    @set_service_action_metadata(endpoint=PromptIdUpdateEndpoint)
    def update(
        self,
        id: str,
        *,
        name: str,
        model_id: str,
        description: Optional[str] = None,
        input: Optional[str] = None,
        output: Optional[str],
        task_id: Optional[str] = None,
        type: Optional[EnumLike[PromptType]] = None,
        messages: Optional[list[ModelLike[BaseMessage]]] = None,
        moderations: Optional[ModelLike[ModerationParameters]] = None,
        parameters: Optional[ModelLike[TextGenerationParameters]] = None,
        data: Optional[ModelLike[PromptTemplateData]] = None,
    ) -> PromptIdUpdateResponse:
        """
        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
            ValidationError: In case of provided parameters are invalid.
        """
        assert_is_not_empty_string(id)
        request_body = _PromptIdUpdateRequest(
            name=name,
            model_id=model_id,
            messages=[to_model_instance(msg, BaseMessage) for msg in messages] if messages else None,
            task_id=task_id,
            description=description,
            moderations=to_model_optional(moderations, ModerationParameters),
            data=to_model_optional(data, PromptTemplateData),
            input=input,
            output=output,
            parameters=to_model_optional(parameters, TextGenerationParameters),
            type=to_enum_optional(type, PromptType),
        ).model_dump()

        self._log_method_execution("Prompts Update", **request_body)

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.update)
            response = client.post(
                url=self._get_endpoint(metadata.endpoint, id=id),
                params=_PromptIdUpdateParametersQuery().model_dump(),
                json=request_body,
            )
            return PromptIdUpdateResponse(**response.json())

    @set_service_action_metadata(endpoint=PromptRetrieveEndpoint)
    def list(
        self,
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        search: Optional[str] = None,
        task_id: Optional[Union[str, list[str]]] = None,
        model_id: Optional[Union[str, list[str]]] = None,
        source: Optional[EnumLikeOrEnumLikeList[PromptRetrieveRequestParamsSource]] = None,
    ) -> PromptRetrieveResponse:
        """
        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
            ValidationError: In case of provided parameters are invalid.
        """
        request_parameters = _PromptRetrieveParametersQuery(
            limit=limit,
            offset=offset,
            search=search,
            task_id=task_id,
            model_id=model_id,
            source=[to_enum(PromptRetrieveRequestParamsSource, s) for s in cast_list(source)] if source else None,
        ).model_dump()
        self._log_method_execution("Prompts List", **request_parameters)

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.list)
            response = client.get(url=self._get_endpoint(metadata.endpoint), params=request_parameters)
            return PromptRetrieveResponse(**response.json())

    @set_service_action_metadata(endpoint=PromptIdDeleteEndpoint)
    def delete(
        self,
        id: str,
    ) -> None:
        """
        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
            ValidationError: In case of provided parameters are invalid.
        """
        assert_is_not_empty_string(id)
        self._log_method_execution("Prompts Delete", id=id)

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.delete)
            client.delete(
                url=self._get_endpoint(metadata.endpoint, id=id),
                params=_PromptIdDeleteParametersQuery().model_dump(),
            )



import sys
from asyncio import Lock, Semaphore, get_running_loop


class _LoopPlaceholder:
    pass


__all__ = ["AsyncioLock", "AsyncioSemaphore"]


class _AsyncioLoopMixin:
    """
    In asyncio prior Python 3.10 the loop being set directly in the constructor.
    Mentioned behaviour is not therefore compatible with our 'SharedEventLoop' approach.
    This class solves that by setting loop once actually needed and not inside the constructor itself.
    """

    _loop = None

    def _get_loop(self):
        if isinstance(self._loop, _LoopPlaceholder):
            self._loop = None

        loop = get_running_loop()

        if self._loop is None:
            self._loop = loop

        if loop is not self._loop:
            raise RuntimeError(f"{self!r} is bound to a different event loop")

        return loop

    def _fix_loop(self):
        if isinstance(self._loop, _LoopPlaceholder) or self._loop is None:
            self._loop = self._get_loop()

    async def acquire(self, *args, **kwargs):
        self._fix_loop()
        await super().acquire(*args, **kwargs)  # type: ignore

    def __init__(self, *args, **kwargs):
        if kwargs.get("loop", None) is None:
            kwargs["loop"] = _LoopPlaceholder()

        super().__init__(*args, **kwargs)


if sys.version_info < (3, 10, 0):

    class AsyncioLock(_AsyncioLoopMixin, Lock):
        pass

    class AsyncioSemaphore(_AsyncioLoopMixin, Semaphore):
        pass


else:
    AsyncioLock = Lock
    AsyncioSemaphore = Semaphore



"""Extension for running local simplified inference API compatible with current version of SDK"""

from genai.extensions.localserver.local_api_server import LocalLLMServer
from genai.extensions.localserver.local_base_model import LocalModel

__all__ = ["LocalModel", "LocalLLMServer"]



from pydantic import Field

from genai.schema import TextGenerationParameters, TextTokenizationParameters
from genai.schema._api import _TextGenerationCreateRequest, _TextTokenizationCreateRequest


class TextGenerationCreateRequest(_TextGenerationCreateRequest):
    """Following properties are required for local server"""

    input: str
    model_id: str
    parameters: TextGenerationParameters = Field(default_factory=lambda: TextGenerationParameters())


class TextTokenizationCreateRequest(_TextTokenizationCreateRequest):
    """Following properties are required for local server"""

    model_id: str
    parameters: TextTokenizationParameters = Field(default_factory=lambda: TextTokenizationParameters())


__all__ = ["TextGenerationCreateRequest", "TextTokenizationCreateRequest"]



import asyncio
import logging
from typing import Awaitable, Callable

from pydantic import BaseModel, Field

from genai._utils.asyncio_future import AsyncioLock
from genai._utils.limiters.adjustable_semaphor import AdjustableAsyncSemaphore
from genai._utils.limiters.base_limiter import BaseLimiter

__all__ = ["ExternalLimiter", "ExternalLimiterHandler", "ConcurrencyResponse"]


class ConcurrencyResponse(BaseModel):
    limit: int = Field(..., ge=1, description="Maximum number of concurrent requests")
    remaining: int = Field(..., description="Number of remaining requests before reaching the limit")


ExternalLimiterHandler = Callable[[], Awaitable[ConcurrencyResponse]]


class ExternalLimiter(BaseLimiter):
    """A limiter class that interacts with an external handler to enforce dynamic concurrent limits."""

    limit_block_recheck_internal_ms: int = 100

    def __init__(self, *, handler: ExternalLimiterHandler):
        self._handler = handler
        self._lock = AsyncioLock()
        self._sp = AdjustableAsyncSemaphore(0)
        self._max_limit = 0
        self._synced = False
        self._logger = logging.getLogger(self.__module__)

    def _change_limit(self, new_limit: int):
        old_limit = self._sp.limit

        limit_has_changed = self._sp.change_max_limit(new_limit)
        if limit_has_changed:
            self._logger.debug(f"Changing limit from {old_limit} to {new_limit}")

    async def report_error(self):
        """Decrease current limit by one"""
        if not self._synced:
            return

        new_limit = 1
        self._change_limit(new_limit)

    async def report_success(self):
        """Increase current limit by one in case it has been decreased previously"""
        if not self._synced:
            return

        new_limit = min(self._max_limit, self._sp.limit + 1)
        self._change_limit(new_limit)

    async def acquire(self):
        async with self._lock:
            if not self._synced:
                self._synced = True

                response = await self._handler()
                while response.remaining <= 0:
                    self._logger.warning(
                        "You can't send more requests due to the API concurrency limit. "
                        "There is another running process (probably). "
                        f"Retrying in {self.limit_block_recheck_internal_ms} ms."
                    )
                    await asyncio.sleep(self.limit_block_recheck_internal_ms / 1000)
                    response = await self._handler()

                self._max_limit = response.limit
                self._change_limit(response.remaining)

        await self._sp.acquire()
        return True

    def release(self):
        self._sp.release()



from abc import ABC, abstractmethod

from genai.schema import (
    TextGenerationParameters,
    TextGenerationResult,
    TextTokenizationCreateResults,
    TextTokenizationParameters,
)

__all__ = ["LocalModel"]


class LocalModel(ABC):
    @property
    def model_id(self):
        """Model ID
        This is the ID that you would use when defining the model you want to use

        example: "google/flan-t5-base"

        Raises:
            NotImplementedError: If you do not implement this property.
        """
        raise NotImplementedError

    @abstractmethod
    def generate(self, input_text: str, parameters: TextGenerationParameters) -> TextGenerationResult:
        """Generate a response from your llm using the provided input text and parameters

        Args:
            input_text: The input prompt chat
            parameters: The parameters that the user code wishes to be used

        Raises:
            NotImplementedError: If you do not implement this function.

        Returns:
            TextGenerationResult: The result to be sent back to the client
        """
        raise NotImplementedError

    @abstractmethod
    def tokenize(self, input_text: str, parameters: TextTokenizationParameters) -> TextTokenizationCreateResults:
        """Tokenize the input text with your model and return the output

        Args:
            input_text (str): The input prompt chat
            parameters (TextTokenizationParameters): The parameter that the user code wishes to be used

        Raises:
            NotImplementedError: If you do not implement this function.

        Returns:
            TextTokenizationCreateResults: The result to be sent back to the client
        """
        raise NotImplementedError



from typing import Annotated, Callable, Optional, TypeVar

from pydantic import Field

__all__ = ["CommonExecutionOptions"]


T = TypeVar("T")


class CommonExecutionOptions:
    __slots__ = ()

    throw_on_error = Annotated[
        bool,
        Field(
            description="Flag indicating whether to throw an error if any error occurs during execution (if disabled, "
            "'None' may be returned in case of error).",
        ),
    ]
    ordered = Annotated[
        bool,
        Field(
            description="Items will be yielded in the order they were passed in, although "
            "they may be processed on the server in different order.",
        ),
    ]
    concurrency_limit = Annotated[
        Optional[int],
        Field(
            description="Upper bound for concurrent executions (in case the passed value is higher than "
            "the API allows, the API's limit will be used).",
            ge=1,
        ),
    ]
    batch_size = Annotated[
        Optional[int],
        Field(
            description="Upper limit for size of single batch of prompts "
            "(the size can be actually lower in case the payload is large enough).",
        ),
    ]
    rate_limit_options = Annotated[
        Optional[dict],
        Field(
            description="HTTPX Transport Options to limit number of requests per second.",
        ),
    ]
    callback = Annotated[
        Optional[Callable[[T], None]],
        Field(description="Callback which is called everytime the response comes."),
    ]



from abc import ABC, abstractmethod


class BaseLimiter(ABC):
    @abstractmethod
    async def acquire(self):
        ...

    @abstractmethod
    def release(self):
        ...

    @abstractmethod
    async def report_error(self):
        ...

    @abstractmethod
    async def report_success(self):
        ...

    async def __aenter__(self):
        await self.acquire()
        # We have no use for the "as ..."  clause in the with
        # statement for locks.
        return None

    async def __aexit__(self, exc_type, exc, tb):
        self.release()



import asyncio
import time
from abc import ABC
from enum import Enum
from typing import Optional, Sequence, Union

from httpx import (
    AsyncHTTPTransport,
    HTTPError,
    HTTPStatusError,
    HTTPTransport,
    Request,
    RequestError,
    Response,
    codes,
)

from genai.exceptions import ApiResponseException

__all__ = ["RetryTransport", "AsyncRetryTransport"]

from genai.exceptions import ApiNetworkException


class BaseRetryTransport(ABC):
    """
    BaseRetryTransport is an abstract base class that provides the basic functionality
    for a retry transport in an API client.

    Attributes:
        Callback: An enumeration of callback types for retry handling.

            - Success: Indicates a successful request.
            - Retry: Indicates a retry attempt.
            - ThresholdReached: Indicates that the retry threshold has been reached.

        default_retry_status_codes (list[int]): The default list of HTTP status codes that should trigger a retry.

    Notes:
        Automatically validates response HTTP status codes.

    """

    class Callback(str, Enum):
        Success = "success"
        Retry = "on_retry"
        ThresholdReached = "on_threshold_reached"

    default_retry_status_codes = [
        codes.TOO_MANY_REQUESTS,
        codes.BAD_GATEWAY,
        codes.SERVICE_UNAVAILABLE,
        codes.CONFLICT,
        codes.TOO_EARLY,
        codes.GATEWAY_TIMEOUT,
        codes.REQUEST_TIMEOUT,
        codes.INTERNAL_SERVER_ERROR,
    ]

    def __init__(
        self,
        *args,
        retry_status_codes: Optional[Sequence[int]] = None,
        backoff_factor: float = 0.2,
        **kwargs,
    ):
        """
        Raises:
            ValueError: If retry_status_codes is provided but is not a list.
        """
        if retry_status_codes is not None and not isinstance(retry_status_codes, list):
            raise ValueError('The "retry_status_code" parameter should be list of status codes!')

        self._retry_status_codes: list[int] = (
            self.default_retry_status_codes if retry_status_codes is None else retry_status_codes
        )
        self._backoff_factor = backoff_factor
        self.retries = kwargs.get("retries", 3)
        super().__init__(*args, **kwargs)

    def _get_execution_plan(self):
        yield 0
        for idx in range(self.retries):
            yield self._backoff_factor * (2**idx)

    def _is_json_response(self, response: Optional[Response]) -> bool:
        return response is not None and "application/json" in response.headers.get("Content-Type", "")

    def _create_exception(
        self, *, exception: Exception, request: Request, requests_count: int
    ) -> Union[ApiResponseException, ApiNetworkException]:
        if isinstance(exception, HTTPStatusError):
            message = (
                f"Failed to handle request after {requests_count - 1} retries to {request.url}."
                if requests_count > 1
                else f"Failed to handle request to {request.url}."
            )

            return ApiResponseException.from_http_response(
                message=message,
                response=exception.response,
            )
        else:
            return ApiNetworkException(f"Network error has occurred (target {request.url}).")


class RetryTransport(BaseRetryTransport, HTTPTransport):
    """
    RetryTransport class handles retrying of HTTP requests with configurable delays and error handling.
    """

    def handle_request(
        self,
        request: Request,
    ) -> Response:
        """
        Raises:
            HTTPStatusError: If the response status code is not in the retry status codes.
            HTTPError: If an HTTP error occurs.
        """
        latest_err: Optional[HTTPError] = None
        requests_count: int = 0

        for delay in self._get_execution_plan():
            requests_count += 1
            if delay > 0:
                time.sleep(delay)

            try:
                response = super().handle_request(request)
                response.request = request
                response.raise_for_status()
                self._execute_callback(self.Callback.Success, request=request, callback_args=[response])
                return response
            except HTTPStatusError as e:
                latest_err = e
                e.response.read()

                if e.response.status_code not in self._retry_status_codes:
                    break

                self._execute_callback(self.Callback.Retry, request=request, callback_args=[e])
            except HTTPError as e:
                latest_err = e
                self._execute_callback(self.Callback.Retry, request=request, callback_args=[e])

        self._execute_callback(self.Callback.ThresholdReached, request=request, callback_args=[])
        raise self._create_exception(
            exception=latest_err, request=request, requests_count=requests_count
        ) from latest_err

    def _execute_callback(self, name: BaseRetryTransport.Callback, request: Request, callback_args: list):
        callback = request.extensions.get(name)
        if callback:
            callback(*callback_args)


class AsyncRetryTransport(BaseRetryTransport, AsyncHTTPTransport):
    """
    Transport responsible for handling asynchronous requests with built-in retry logic.
    """

    async def handle_async_request(
        self,
        request: Request,
    ) -> Response:
        """
        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
        """
        latest_err: Optional[HTTPError] = None
        requests_count: int = 0

        for delay in self._get_execution_plan():
            requests_count += 1
            if delay > 0:
                await asyncio.sleep(delay)

            try:
                response = await super().handle_async_request(request)
                response.request = request
                response.raise_for_status()
                await self._execute_async_callback(self.Callback.Success, request=request, callback_args=[response])
                return response
            except HTTPStatusError as e:
                latest_err = e
                await e.response.aread()

                if e.response.status_code not in self._retry_status_codes:
                    break

                await self._execute_async_callback(self.Callback.Retry, request=request, callback_args=[e])
            except RequestError as e:
                latest_err = e
                await self._execute_async_callback(self.Callback.Retry, request=request, callback_args=[e])

        await self._execute_async_callback(self.Callback.ThresholdReached, request=request, callback_args=[])
        raise self._create_exception(
            exception=latest_err, request=request, requests_count=requests_count
        ) from latest_err

    async def _execute_async_callback(self, name: BaseRetryTransport.Callback, request: Request, callback_args: list):
        callback = request.extensions.get(name)
        if callback:
            await callback(*callback_args)



from __future__ import annotations

from typing import Any, Optional, TypeVar


def assert_is_not_empty_string(input: Any):
    if not isinstance(input, str):
        raise TypeError(f"Provided value '{input}' is not a string!")
    if input == "":
        raise ValueError("Provided value must be non empty!")


T = TypeVar("T")


def assert_is_instanceof(input: Optional[T], target: type[T]):
    assert isinstance(input, target)
    if not input or not isinstance(input, target):
        raise ValueError(f"Provided value '{input}' is not instance of {target.__name__}!")



from __future__ import annotations

from pathlib import Path
from typing import Any, Mapping, Optional, Union

from langchain_core.outputs import ChatGenerationChunk, GenerationChunk
from pydantic import BaseModel

from genai._utils.general import merge_objects
from genai.extensions._common.utils import extract_token_usage
from genai.schema import TextGenerationParameters


def update_token_usage(*, target: dict[str, Any], source: Optional[dict[str, Any]]):
    if not source:
        return

    for key, value in source.items():
        if key in target:
            target[key] += value
        else:
            target[key] = value


def update_token_usage_stream(*, target: dict[str, Any], source: Optional[dict]):
    if not source:
        return

    def get_value(key: str, override=False) -> int:
        current = target.get(key, 0) or 0
        new = source.get(key, 0) or 0

        if new != 0 and (current == 0 or override):
            return new
        else:
            return current

    completion_tokens = get_value("completion_tokens", override=True)
    prompt_tokens = get_value("prompt_tokens")
    target.update(
        {
            "prompt_tokens": prompt_tokens,
            "input_token_count": prompt_tokens,
            "completion_tokens": completion_tokens,
            "total_tokens": completion_tokens + prompt_tokens,
            "generated_token_count": completion_tokens,
        }
    )


def create_llm_output(*, model: str, token_usages: Optional[list[Optional[dict]]] = None, **kwargs) -> dict[str, Any]:
    final_token_usage = extract_token_usage({})
    for source in token_usages or []:
        update_token_usage(target=final_token_usage, source=source)
    return {"model_name": model, "token_usage": final_token_usage, **kwargs}


def load_config(file: Union[str, Path]) -> dict:
    def parse_config() -> dict:
        file_path = Path(file) if isinstance(file, str) else file
        if file_path.suffix == ".json":
            with open(file_path) as f:
                import json

                return json.load(f)
        elif file_path.suffix == ".yaml":
            with open(file_path, "r") as f:
                import yaml

                return yaml.safe_load(f)
        else:
            raise ValueError("File type must be json or yaml")

    config = parse_config()
    config["parameters"] = TextGenerationParameters(**config.get("parameters", {}))
    return config


def dump_optional_model(model: Optional[BaseModel]) -> Optional[Mapping[str, Any]]:
    return model.model_dump(exclude_none=True) if model else None


class CustomChatGenerationChunk(ChatGenerationChunk):
    def __add__(self, other: ChatGenerationChunk) -> CustomChatGenerationChunk:
        """ "Replaces LangChain's 'merge_dicts' with our simplified 'merge_objects' utility"""
        if isinstance(other, ChatGenerationChunk):
            generation_info = merge_objects(
                self.generation_info or {},
                other.generation_info or {},
            )
            return CustomChatGenerationChunk(
                message=self.message + other.message,
                generation_info=generation_info or None,
            )
        else:
            raise TypeError(f"unsupported operand type(s) for +: '{type(self)}' and '{type(other)}'")


class CustomGenerationChunk(GenerationChunk):
    def __add__(self, other: GenerationChunk) -> CustomGenerationChunk:
        """Replaces LangChain's 'merge_dicts' with our simplified 'merge_objects' utility"""

        if isinstance(other, GenerationChunk):
            generation_info = merge_objects(
                self.generation_info or {},
                other.generation_info or {},
            )
            return CustomGenerationChunk(
                text=self.text + other.text,
                generation_info=generation_info or None,
            )
        else:
            raise TypeError(f"unsupported operand type(s) for +: '{type(self)}' and '{type(other)}'")



from __future__ import annotations

import logging
import re
from abc import ABC
from enum import Enum
from typing import Generic, Optional, TypeVar, Union, cast
from urllib.parse import quote

from pydantic import BaseModel, ConfigDict

from genai._utils.api_client import ApiClient
from genai._utils.general import to_model_instance
from genai._utils.http_client.httpx_client import AsyncHttpxClient, HttpxClient
from genai._utils.service.metadata import inherit_metadata
from genai._utils.shared_options import CommonExecutionOptions
from genai._utils.validators import assert_is_not_empty_string
from genai.schema._endpoints import ApiEndpoint

__all__ = ["BaseService", "BaseServiceConfig", "BaseServiceServices", "CommonExecutionOptions"]


class BaseServiceConfig(BaseModel):
    """Options which given service uses"""

    model_config = ConfigDict(extra="forbid", validate_assignment=True, validate_default=True)


class BaseServiceServices(BaseModel):
    """List of services which given service uses"""

    model_config = ConfigDict(extra="forbid", validate_assignment=True, validate_default=True)


TConfig = TypeVar("TConfig", bound=BaseServiceConfig)
TServices = TypeVar("TServices", bound=BaseServiceServices)

TClient = TypeVar("TClient", HttpxClient, AsyncHttpxClient)


class BaseService(Generic[TConfig, TServices], ABC):
    """
    A base class for services that interact with an API.

    Attributes:
        Config (type): The type of the configuration class used by the service.
        Services (type): The type of the services class used by the service.
    """

    Config: type[TConfig] = cast(type[TConfig], BaseServiceConfig)
    Services: type[TServices] = cast(type[TServices], BaseServiceServices)

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        for name in dir(cls):
            method = getattr(cls, name, None)
            if not callable(method) or name.startswith("__"):
                continue

            for base_cls in cls.__bases__:
                if not issubclass(base_cls, BaseService):
                    continue

                target = getattr(base_cls, name, None)
                if not target:
                    continue

                inherit_metadata(source=target, target=method)

    def __init__(
        self,
        *,
        api_client: ApiClient,
        config: Optional[Union[TConfig, dict]] = None,
    ):
        self._logger = logging.getLogger(f"{self.__module__}")
        self._api_client = api_client
        self.config = to_model_instance(config, self.Config)

    def _get_http_client(self, **kwargs) -> HttpxClient:
        return self._api_client.get_http_client(**kwargs)

    def _get_async_http_client(self, **kwargs):
        return self._api_client.get_async_http_client(**kwargs)

    def _log_method_execution(self, name: str, /, **kwargs):
        self._logger.debug("Calling %s with params: %s", name, kwargs)

    @staticmethod
    def _get_endpoint(
        endpoint: type[ApiEndpoint],
        **params: Union[int, str, Enum],
    ) -> str:
        target_endpoint = endpoint.path
        if not target_endpoint:
            raise ValueError("Endpoint was not found in the provided config.")

        for k, v in params.items():
            v = v.value if isinstance(v, Enum) else str(v)
            assert_is_not_empty_string(v)
            parameter_expression = f"{{{k}}}"
            if parameter_expression not in target_endpoint:
                raise ValueError(f"Variable '{k}' is not inside provided endpoint '{target_endpoint}'!")
            v = quote(v, safe=[])
            target_endpoint = target_endpoint.replace(parameter_expression, v)

        missing_path_parameters = re.findall(r"\{.*?}", target_endpoint)
        if missing_path_parameters:
            raise ValueError(
                f"Missing {','.join(missing_path_parameters)} parameters for '{target_endpoint}' endpoint."
            )

        return target_endpoint



import asyncio
from typing import Optional

from genai._types import ModelLike
from genai.client import Client
from genai.schema import TextEmbeddingParameters
from genai.text.embedding.embedding_service import CreateExecutionOptions

try:
    from llama_index.embeddings.base import BaseEmbedding, Embedding
except ImportError:
    raise ImportError("Could not import llamaindex: Please install ibm-generative-ai[llama-index] extension.")  # noqa: B904


class IBMGenAILlamaIndexEmbedding(BaseEmbedding):
    client: Client
    model_id: str
    parameters: Optional[ModelLike[TextEmbeddingParameters]] = None
    execution_options: Optional[ModelLike[CreateExecutionOptions]] = None

    @classmethod
    def class_name(cls) -> str:
        return "IBMGenAIEmbedding"

    def _get_query_embedding(self, query: str) -> Embedding:
        response = self._get_embeddings([query])
        return response[0]

    def _get_text_embedding(self, text: str) -> Embedding:
        response = self._get_embeddings([text])
        return response[0]

    def _get_text_embeddings(self, texts: list[str]) -> list[Embedding]:
        response = self._get_embeddings(texts)
        return response

    async def _aget_query_embedding(self, query: str) -> Embedding:
        return await asyncio.get_running_loop().run_in_executor(None, self._get_query_embedding, query)

    async def _aget_text_embedding(self, text: str) -> Embedding:
        return await asyncio.get_running_loop().run_in_executor(None, self._get_text_embedding, text)

    async def _aget_text_embeddings(self, texts: list[str]) -> list[Embedding]:
        return await asyncio.get_running_loop().run_in_executor(None, self._get_text_embeddings, texts)

    def _get_embeddings(self, texts: list[str]) -> list[Embedding]:
        embeddings: list[list[float]] = []
        for response in self.client.text.embedding.create(
            model_id=self.model_id, inputs=texts, parameters=self.parameters, execution_options=self.execution_options
        ):
            embeddings.extend(response.results)

        return embeddings



class ApiEndpoint:
    __slots__ = ()
    path: str
    method: str
    version: str
    class_name: str


class ApiKeyRetrieveEndpoint(ApiEndpoint):
    path: str = "/v2/api_key"
    method: str = "GET"
    version: str = "2023-11-22"


class ApiKeyRegenerateCreateEndpoint(ApiEndpoint):
    path: str = "/v2/api_key/regenerate"
    method: str = "POST"
    version: str = "2023-11-22"


class FileRetrieveEndpoint(ApiEndpoint):
    path: str = "/v2/files"
    method: str = "GET"
    version: str = "2023-12-15"


class FileCreateEndpoint(ApiEndpoint):
    path: str = "/v2/files"
    method: str = "POST"
    version: str = "2023-12-15"


class FileIdDeleteEndpoint(ApiEndpoint):
    path: str = "/v2/files/{id}"
    method: str = "DELETE"
    version: str = "2023-11-22"


class FileIdRetrieveEndpoint(ApiEndpoint):
    path: str = "/v2/files/{id}"
    method: str = "GET"
    version: str = "2023-12-15"


class FileIdContentRetrieveEndpoint(ApiEndpoint):
    path: str = "/v2/files/{id}/content"
    method: str = "GET"
    version: str = "2023-11-22"


class ModelRetrieveEndpoint(ApiEndpoint):
    path: str = "/v2/models"
    method: str = "GET"
    version: str = "2023-11-22"


class ModelIdRetrieveEndpoint(ApiEndpoint):
    path: str = "/v2/models/{id}"
    method: str = "GET"
    version: str = "2024-01-30"


class PromptRetrieveEndpoint(ApiEndpoint):
    path: str = "/v2/prompts"
    method: str = "GET"
    version: str = "2024-01-10"


class PromptCreateEndpoint(ApiEndpoint):
    path: str = "/v2/prompts"
    method: str = "POST"
    version: str = "2024-01-10"


class PromptIdDeleteEndpoint(ApiEndpoint):
    path: str = "/v2/prompts/{id}"
    method: str = "DELETE"
    version: str = "2023-11-22"


class PromptIdRetrieveEndpoint(ApiEndpoint):
    path: str = "/v2/prompts/{id}"
    method: str = "GET"
    version: str = "2024-01-10"


class PromptIdPatchEndpoint(ApiEndpoint):
    path: str = "/v2/prompts/{id}"
    method: str = "PATCH"
    version: str = "2024-01-10"


class PromptIdUpdateEndpoint(ApiEndpoint):
    path: str = "/v2/prompts/{id}"
    method: str = "PUT"
    version: str = "2024-01-10"


class RequestRetrieveEndpoint(ApiEndpoint):
    path: str = "/v2/requests"
    method: str = "GET"
    version: str = "2023-11-22"


class RequestChatConversationIdDeleteEndpoint(ApiEndpoint):
    path: str = "/v2/requests/chat/{conversationId}"
    method: str = "DELETE"
    version: str = "2023-11-22"


class RequestChatConversationIdRetrieveEndpoint(ApiEndpoint):
    path: str = "/v2/requests/chat/{conversationId}"
    method: str = "GET"
    version: str = "2023-11-22"


class RequestIdDeleteEndpoint(ApiEndpoint):
    path: str = "/v2/requests/{id}"
    method: str = "DELETE"
    version: str = "2023-11-22"


class SystemPromptRetrieveEndpoint(ApiEndpoint):
    path: str = "/v2/system_prompts"
    method: str = "GET"
    version: str = "2023-11-22"


class SystemPromptCreateEndpoint(ApiEndpoint):
    path: str = "/v2/system_prompts"
    method: str = "POST"
    version: str = "2023-11-22"


class SystemPromptIdDeleteEndpoint(ApiEndpoint):
    path: str = "/v2/system_prompts/{id}"
    method: str = "DELETE"
    version: str = "2023-11-22"


class SystemPromptIdRetrieveEndpoint(ApiEndpoint):
    path: str = "/v2/system_prompts/{id}"
    method: str = "GET"
    version: str = "2023-11-22"


class SystemPromptIdUpdateEndpoint(ApiEndpoint):
    path: str = "/v2/system_prompts/{id}"
    method: str = "PUT"
    version: str = "2023-11-22"


class TaskRetrieveEndpoint(ApiEndpoint):
    path: str = "/v2/tasks"
    method: str = "GET"
    version: str = "2023-11-22"


class TextChatCreateEndpoint(ApiEndpoint):
    path: str = "/v2/text/chat"
    method: str = "POST"
    version: str = "2024-01-10"


class TextChatOutputCreateEndpoint(ApiEndpoint):
    path: str = "/v2/text/chat/output"
    method: str = "POST"
    version: str = "2024-01-10"


class TextChatStreamCreateEndpoint(ApiEndpoint):
    path: str = "/v2/text/chat_stream"
    method: str = "POST"
    version: str = "2024-01-10"


class TextEmbeddingCreateEndpoint(ApiEndpoint):
    path: str = "/v2/text/embeddings"
    method: str = "POST"
    version: str = "2023-11-22"


class TextEmbeddingLimitRetrieveEndpoint(ApiEndpoint):
    path: str = "/v2/text/embeddings/limits"
    method: str = "GET"
    version: str = "2023-11-22"


class TextExtractionLimitRetrieveEndpoint(ApiEndpoint):
    path: str = "/v2/text/extraction/limits"
    method: str = "GET"
    version: str = "2023-11-22"


class TextGenerationCreateEndpoint(ApiEndpoint):
    path: str = "/v2/text/generation"
    method: str = "POST"
    version: str = "2024-01-10"


class TextGenerationComparisonCreateEndpoint(ApiEndpoint):
    path: str = "/v2/text/generation/comparison"
    method: str = "POST"
    version: str = "2023-11-22"


class TextGenerationLimitRetrieveEndpoint(ApiEndpoint):
    path: str = "/v2/text/generation/limits"
    method: str = "GET"
    version: str = "2023-11-22"


class TextGenerationOutputCreateEndpoint(ApiEndpoint):
    path: str = "/v2/text/generation/output"
    method: str = "POST"
    version: str = "2023-11-22"


class TextGenerationIdFeedbackRetrieveEndpoint(ApiEndpoint):
    path: str = "/v2/text/generation/{id}/feedback"
    method: str = "GET"
    version: str = "2023-11-22"


class TextGenerationIdFeedbackCreateEndpoint(ApiEndpoint):
    path: str = "/v2/text/generation/{id}/feedback"
    method: str = "POST"
    version: str = "2023-11-22"


class TextGenerationIdFeedbackUpdateEndpoint(ApiEndpoint):
    path: str = "/v2/text/generation/{id}/feedback"
    method: str = "PUT"
    version: str = "2023-11-22"


class TextGenerationStreamCreateEndpoint(ApiEndpoint):
    path: str = "/v2/text/generation_stream"
    method: str = "POST"
    version: str = "2024-01-10"


class TextModerationCreateEndpoint(ApiEndpoint):
    path: str = "/v2/text/moderations"
    method: str = "POST"
    version: str = "2023-11-22"


class TextTokenizationCreateEndpoint(ApiEndpoint):
    path: str = "/v2/text/tokenization"
    method: str = "POST"
    version: str = "2024-01-10"


class TuneRetrieveEndpoint(ApiEndpoint):
    path: str = "/v2/tunes"
    method: str = "GET"
    version: str = "2023-11-22"


class TuneCreateEndpoint(ApiEndpoint):
    path: str = "/v2/tunes"
    method: str = "POST"
    version: str = "2023-11-22"


class TuneImportCreateEndpoint(ApiEndpoint):
    path: str = "/v2/tunes/import"
    method: str = "POST"
    version: str = "2023-11-22"


class TuneIdDeleteEndpoint(ApiEndpoint):
    path: str = "/v2/tunes/{id}"
    method: str = "DELETE"
    version: str = "2023-11-22"


class TuneIdRetrieveEndpoint(ApiEndpoint):
    path: str = "/v2/tunes/{id}"
    method: str = "GET"
    version: str = "2023-11-22"


class TuneIdPatchEndpoint(ApiEndpoint):
    path: str = "/v2/tunes/{id}"
    method: str = "PATCH"
    version: str = "2023-11-22"


class TuneIdContentTypeRetrieveEndpoint(ApiEndpoint):
    path: str = "/v2/tunes/{id}/content/{type}"
    method: str = "GET"
    version: str = "2023-12-15"


class TuningTypeRetrieveEndpoint(ApiEndpoint):
    path: str = "/v2/tuning_types"
    method: str = "GET"
    version: str = "2024-01-30"


class UserDeleteEndpoint(ApiEndpoint):
    path: str = "/v2/user"
    method: str = "DELETE"
    version: str = "2023-11-22"


class UserRetrieveEndpoint(ApiEndpoint):
    path: str = "/v2/user"
    method: str = "GET"
    version: str = "2023-11-22"


class UserPatchEndpoint(ApiEndpoint):
    path: str = "/v2/user"
    method: str = "PATCH"
    version: str = "2023-11-22"


class UserCreateEndpoint(ApiEndpoint):
    path: str = "/v2/user"
    method: str = "POST"
    version: str = "2023-11-22"



import asyncio
from asyncio import CancelledError, Future
from collections import deque
from contextlib import suppress

from genai._utils.asyncio_future import AsyncioSemaphore


class AdjustableAsyncSemaphore(AsyncioSemaphore):
    """
    AdjustableAsyncSemaphore is a subclass of Semaphore that allows for dynamic adjustment of its maximum limit.
    It includes methods for changing the maximum limit and getting information about the current state of the semaphore.
    """

    def __init__(self, value: int = 1):
        super().__init__(value)
        self._max_limit = value
        self._waiters = deque()
        self._processing = 0
        self._dummy_waiters: set[Future] = set()

    @property
    def limit(self):
        return self._max_limit

    @property
    def processing(self):
        return self._processing

    @property
    def waiting(self):
        if self._waiters is None:
            return 0

        return len(self._waiters) - len(self._dummy_waiters)

    async def acquire(self):
        """Taken from Python 3.11"""

        if not self.locked():
            self._value -= 1
            self._processing += 1
            return True

        fut = self._get_loop().create_future()
        self._waiters.append(fut)

        # Finally block should be called before the CancelledError
        # handling as we don't want CancelledError to call
        # _wake_up_first() and attempt to wake up itself.
        try:
            try:
                await fut
            finally:
                self._waiters.remove(fut)
        except CancelledError:
            if not fut.cancelled():
                self._value += 1
                self._processing -= 1
                self._wake_up_next()
            raise

        if self._value > 0:
            self._wake_up_next()
        return True

    def release(self):
        """Taken from Python 3.11"""
        self._value += 1
        self._processing -= 1
        self._wake_up_next()

    def _wake_up_next(self):
        """Taken from Python 3.11"""
        if not self._waiters:
            return

        for fut in self._waiters:
            if not fut.done():
                self._value -= 1
                if fut not in self._dummy_waiters:
                    self._processing += 1
                fut.set_result(True)
                return

    def change_max_limit(self, new_limit: int) -> bool:
        """
        Change the maximum limit of the semaphore.

        Args:
            new_limit (int): The new maximum limit for the semaphore.

        Returns:
            bool: Returns `True` if the maximum limit was changed, `False` otherwise.

        Raises:
            ValueError: If `new_limit` is less than 0.
        """
        if new_limit < 0:
            raise ValueError("Semaphore concurrency cannot be less than 0!")

        if new_limit == self._max_limit:
            return False

        old_limit = self._max_limit
        self._max_limit = new_limit

        if new_limit > old_limit:
            points_to_add = new_limit - old_limit
            for _ in range(0, points_to_add):
                self._processing += 1
                self.release()
        else:
            points_to_remove = old_limit - new_limit

            for _ in range(0, points_to_remove):
                if self.locked():
                    fut: asyncio.Future = self._get_loop().create_future()

                    def handle_done(f: Future):
                        assert f.done()

                        with suppress(ValueError):
                            self._waiters.remove(f)

                        if self._value > 0:
                            self._wake_up_next()

                        with suppress(KeyError):
                            self._dummy_waiters.remove(f)

                    fut.add_done_callback(handle_done)

                    self._dummy_waiters.add(fut)
                    self._waiters.appendleft(fut)
                else:
                    self._value -= 1

        return True



from typing import Generator, Optional, Sequence, Union

from genai._types import EnumLike, ModelLike
from genai._utils.api_client import ApiClient
from genai._utils.general import to_enum_optional, to_model_instance, to_model_optional
from genai._utils.service import (
    BaseService,
    BaseServiceConfig,
    BaseServiceServices,
    get_service_action_metadata,
    set_service_action_metadata,
)
from genai.request.request_service import RequestService as _RequestService
from genai.schema import (
    BaseMessage,
    ModerationParameters,
    TextChatCreateResponse,
    TextChatStreamCreateResponse,
    TextGenerationParameters,
    TrimMethod,
)
from genai.schema._api import (
    _TextChatCreateParametersQuery,
    _TextChatCreateRequest,
    _TextChatStreamCreateRequest,
    _TextGenerationStreamCreateParametersQuery,
)
from genai.schema._endpoints import (
    TextChatCreateEndpoint,
    TextChatStreamCreateEndpoint,
)
from genai.text.generation._generation_utils import generation_stream_handler

__all__ = ["ChatService", "BaseServices"]


class BaseServices(BaseServiceServices):
    RequestService: type[_RequestService] = _RequestService


class ChatService(BaseService[BaseServiceConfig, BaseServices]):
    Services = BaseServices

    def __init__(
        self,
        *,
        api_client: ApiClient,
        config: Optional[Union[BaseServiceConfig, dict]] = None,
        services: Optional[BaseServices] = None,
    ):
        super().__init__(api_client=api_client, config=config)
        if not services:
            services = BaseServices()

        self._request = services.RequestService(api_client=api_client)

    @set_service_action_metadata(endpoint=TextChatCreateEndpoint)
    def create(
        self,
        *,
        conversation_id: Optional[str] = None,
        model_id: Optional[str] = None,
        messages: Optional[Sequence[BaseMessage]] = None,
        moderations: Optional[ModelLike[ModerationParameters]] = None,
        parameters: Optional[ModelLike[TextGenerationParameters]] = None,
        parent_id: Optional[str] = None,
        prompt_id: Optional[str] = None,
        prompt_template_id: Optional[str] = None,
        trim_method: Optional[EnumLike[TrimMethod]] = None,
        use_conversation_parameters: Optional[bool] = None,
    ) -> TextChatCreateResponse:
        """
        Example::

            from genai import Client, Credentials
            from genai.text.chat import HumanMessage, TextGenerationParameters

            client = Client(credentials=Credentials.from_env())

            # Create a new conversation
            response = client.text.chat.create(
                model_id="meta-llama/llama-2-70b-chat",
                messages=[HumanMessage(content="Describe the game Chess?")],
                parameters=TextGenerationParameters(max_token_limit=100)
            )
            conversation_id = response.conversation_id
            print(f"Response: {response.results[0].generated_text}")

            # Continue in the conversation
            response = client.text.chat.create(
                conversation_id=conversation_id,
                use_conversation_parameters=True,
                messages=[HumanMessage(content="Who is the best player of that game?")]
            )
            print(f"Response: {response.results[0].generated_text}")

        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
            ValidationError: In case of provided parameters are invalid.
        """
        metadata = get_service_action_metadata(self.create)
        request_body = _TextChatCreateRequest(
            model_id=model_id,
            conversation_id=conversation_id,
            messages=self._prepare_messages(messages) if messages is not None else None,
            moderations=to_model_optional(moderations, ModerationParameters),
            parameters=to_model_optional(parameters, TextGenerationParameters),
            parent_id=parent_id,
            prompt_id=prompt_id,
            prompt_template_id=prompt_template_id,
            trim_method=to_enum_optional(trim_method, TrimMethod),
            use_conversation_parameters=use_conversation_parameters or None,
        ).model_dump()

        self._log_method_execution("Text Chat Generation Create", **request_body)

        with self._get_http_client() as client:
            http_response = client.post(
                url=self._get_endpoint(metadata.endpoint),
                params=_TextChatCreateParametersQuery().model_dump(),
                json=request_body,
            )
            return TextChatCreateResponse(**http_response.json())

    @set_service_action_metadata(endpoint=TextChatStreamCreateEndpoint)
    def create_stream(
        self,
        *,
        model_id: Optional[str] = None,
        conversation_id: Optional[str] = None,
        messages: Optional[list[BaseMessage]] = None,
        moderations: Optional[ModelLike[ModerationParameters]] = None,
        parameters: Optional[ModelLike[TextGenerationParameters]] = None,
        parent_id: Optional[str] = None,
        prompt_id: Optional[str] = None,
        prompt_template_id: Optional[str] = None,
        trim_method: Optional[EnumLike[TrimMethod]] = None,
        use_conversation_parameters: Optional[bool] = None,
    ) -> Generator[TextChatStreamCreateResponse, None, None]:
        """
        Example::

            from genai import Client, Credentials
            from genai.text.chat import HumanMessage, TextGenerationParameters

            client = Client(credentials=Credentials.from_env())

            # Create a new conversation
            for response in client.text.chat.create_stream(
                    model_id="meta-llama/llama-2-70b-chat",
                    messages=[HumanMessage(content="Describe the game Chess?")],
                    parameters=TextGenerationParameters(max_token_limit=100)
                ):
                print(f"Chunk retrieved: {response.results[0].generated_text}")

        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
            ValidationError: In case of provided parameters are invalid.
        """
        metadata = get_service_action_metadata(self.create_stream)
        request_body = _TextChatStreamCreateRequest(
            model_id=model_id,
            conversation_id=conversation_id,
            messages=self._prepare_messages(messages) if messages is not None else None,
            moderations=to_model_optional(moderations, ModerationParameters),
            parameters=to_model_optional(parameters, TextGenerationParameters),
            parent_id=parent_id,
            prompt_id=prompt_id,
            prompt_template_id=prompt_template_id,
            trim_method=to_enum_optional(trim_method, TrimMethod),
            use_conversation_parameters=use_conversation_parameters or None,
        ).model_dump()

        self._log_method_execution("Text Chat Generation Create Stream", **request_body)

        with self._get_http_client() as client:
            yield from generation_stream_handler(
                ResponseModel=TextChatStreamCreateResponse,
                logger=self._logger,
                generator=client.post_stream(
                    url=self._get_endpoint(metadata.endpoint),
                    params=_TextGenerationStreamCreateParametersQuery().model_dump(),
                    json=request_body,
                ),
            )

    def _prepare_messages(self, messages: Sequence[Union[BaseMessage, dict]]) -> list[BaseMessage]:
        return [to_model_instance(msg, BaseMessage) for msg in messages]



from typing import Generator, Optional, Union

import httpx
from httpx import AsyncClient, HTTPStatusError
from pydantic import BaseModel

from genai._types import ModelLike
from genai._utils.api_client import ApiClient
from genai._utils.async_executor import execute_async
from genai._utils.general import cast_list, to_model_instance, to_model_optional
from genai._utils.service import (
    BaseService,
    BaseServiceConfig,
    BaseServiceServices,
    CommonExecutionOptions,
    get_service_action_metadata,
    set_service_action_metadata,
)
from genai.schema import (
    ModerationParameters,
    PromptTemplateData,
    TextGenerationComparisonCreateEndpoint,
    TextGenerationComparisonCreateRequestRequest,
    TextGenerationComparisonCreateResponse,
    TextGenerationComparisonParameters,
    TextGenerationCreateEndpoint,
    TextGenerationCreateResponse,
    TextGenerationParameters,
    TextGenerationStreamCreateEndpoint,
    TextGenerationStreamCreateResponse,
)
from genai.schema._api import (
    _TextGenerationComparisonCreateParametersQuery,
    _TextGenerationComparisonCreateRequest,
    _TextGenerationCreateParametersQuery,
    _TextGenerationCreateRequest,
    _TextGenerationStreamCreateParametersQuery,
    _TextGenerationStreamCreateRequest,
)
from genai.text.generation._generation_utils import generation_stream_handler
from genai.text.generation.feedback.feedback_service import FeedbackService as _FeedbackService
from genai.text.generation.limits.limit_service import LimitService as _LimitService

__all__ = ["GenerationService", "BaseConfig", "BaseServices", "CreateExecutionOptions"]


from genai._utils.http_client.retry_transport import BaseRetryTransport
from genai._utils.limiters.base_limiter import BaseLimiter
from genai._utils.limiters.external_limiter import ConcurrencyResponse, ExternalLimiter
from genai._utils.limiters.local_limiter import LocalLimiter
from genai._utils.limiters.shared_limiter import LoopBoundLimiter


class BaseServices(BaseServiceServices):
    LimitService: type[_LimitService] = _LimitService
    FeedbackService: type[_FeedbackService] = _FeedbackService


class CreateExecutionOptions(BaseModel):
    throw_on_error: CommonExecutionOptions.throw_on_error = True
    ordered: CommonExecutionOptions.ordered = True
    concurrency_limit: CommonExecutionOptions.concurrency_limit = None
    callback: CommonExecutionOptions.callback[
        Union[TextGenerationStreamCreateResponse, TextGenerationCreateResponse]
    ] = None


class BaseConfig(BaseServiceConfig):
    create_execution_options: CreateExecutionOptions = CreateExecutionOptions()


class GenerationService(BaseService[BaseConfig, BaseServices]):
    Config = BaseConfig
    Services = BaseServices

    def __init__(
        self,
        *,
        api_client: ApiClient,
        services: Optional[BaseServices] = None,
        config: Optional[Union[BaseConfig, dict]] = None,
    ):
        super().__init__(api_client=api_client, config=config)

        if not services:
            services = BaseServices()

        self._concurrency_limiter = self._get_concurrency_limiter()
        self.feedback = services.FeedbackService(api_client=api_client)
        self.limit = services.LimitService(api_client=api_client)

    def _get_concurrency_limiter(self) -> LoopBoundLimiter:
        async def handler():
            response = await self.limit.aretrieve()
            return ConcurrencyResponse(
                limit=response.result.concurrency.limit,
                remaining=response.result.concurrency.remaining,
            )

        return LoopBoundLimiter(lambda: ExternalLimiter(handler=handler))

    @set_service_action_metadata(endpoint=TextGenerationCreateEndpoint)
    def create(
        self,
        *,
        model_id: Optional[str] = None,
        prompt_id: Optional[str] = None,
        inputs: Optional[Union[list[str], str]] = None,
        parameters: Optional[ModelLike[TextGenerationParameters]] = None,
        moderations: Optional[ModelLike[ModerationParameters]] = None,
        data: Optional[ModelLike[PromptTemplateData]] = None,
        execution_options: Optional[ModelLike[CreateExecutionOptions]] = None,
    ) -> Generator[TextGenerationCreateResponse, None, None]:
        """
        Args:
            model_id: The ID of the model.
            prompt_id: The ID of the prompt which should be used.
            inputs: Prompt/prompts to process. It is recommended not to leave any trailing spaces.
            parameters: Parameters for text generation.
            moderations: Parameters for moderation.
            data: An optional data object for underlying prompt.
            execution_options: An optional configuration how SDK should work (error handling, limits, callbacks, ...)

        Yields:
            TextGenerationCreateResponse object (server response without modification).

        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
            ValidationError: In case of provided parameters are invalid.

        Note:
            To limit number of concurrent requests or change execution procedure, see 'execute_options' parameter.
        """
        metadata = get_service_action_metadata(self.create)
        prompts = cast_list(inputs) if inputs else []
        parameters_formatted = to_model_optional(parameters, TextGenerationParameters)
        moderations_formatted = to_model_optional(moderations, ModerationParameters)
        template_formatted = to_model_optional(data, PromptTemplateData)
        execution_options_formatted = to_model_instance(
            [self.config.create_execution_options, execution_options], CreateExecutionOptions
        )
        assert execution_options_formatted

        self._log_method_execution(
            "Generate Create",
            prompts=prompts,
            prompt_id=prompt_id,
            parameters=parameters_formatted,
            moderations=moderations_formatted,
            data=template_formatted,
            execution_options=execution_options_formatted,
        )

        if prompt_id is not None:
            with self._get_http_client() as client:
                http_response = client.post(
                    url=self._get_endpoint(metadata.endpoint),
                    params=_TextGenerationCreateParametersQuery().model_dump(),
                    json=_TextGenerationCreateRequest(
                        input=None,
                        model_id=model_id,
                        moderations=moderations_formatted,
                        parameters=parameters_formatted,
                        prompt_id=prompt_id,
                        data=template_formatted,
                    ).model_dump(),
                )
                yield TextGenerationCreateResponse(**http_response.json())
                return

        async def handler(input: str, http_client: AsyncClient, limiter: BaseLimiter) -> TextGenerationCreateResponse:
            self._log_method_execution("Generate Create - processing input", input=input)

            async def handle_retry(ex: Exception):
                if isinstance(ex, HTTPStatusError) and ex.response.status_code == httpx.codes.TOO_MANY_REQUESTS:
                    await limiter.report_error()

            async def handle_success(*args):
                await limiter.report_success()

            http_response = await http_client.post(
                url=self._get_endpoint(metadata.endpoint),
                extensions={
                    BaseRetryTransport.Callback.Retry: handle_retry,
                    BaseRetryTransport.Callback.Success: handle_success,
                },
                params=_TextGenerationCreateParametersQuery().model_dump(),
                json=_TextGenerationCreateRequest(
                    input=input,
                    model_id=model_id,
                    moderations=moderations_formatted,
                    parameters=parameters_formatted,
                    prompt_id=prompt_id,
                    data=template_formatted,
                ).model_dump(),
            )
            response = TextGenerationCreateResponse(**http_response.json())

            if execution_options_formatted.callback:
                execution_options_formatted.callback(response)

            return response

        yield from execute_async(
            inputs=prompts,
            handler=handler,
            limiters=[
                self._concurrency_limiter,
                self._get_local_limiter(execution_options_formatted.concurrency_limit),
            ],
            http_client=self._get_async_http_client,
            ordered=execution_options_formatted.ordered,
            throw_on_error=execution_options_formatted.throw_on_error,
        )

    @set_service_action_metadata(endpoint=TextGenerationStreamCreateEndpoint)
    def create_stream(
        self,
        *,
        input: Optional[str] = None,
        model_id: Optional[str] = None,
        prompt_id: Optional[str] = None,
        parameters: Optional[ModelLike[TextGenerationParameters]] = None,
        moderations: Optional[ModelLike[ModerationParameters]] = None,
        data: Optional[ModelLike[PromptTemplateData]] = None,
    ) -> Generator[TextGenerationStreamCreateResponse, None, None]:
        """
        Yields:
            TextGenerationStreamCreateResponse (raw server response object)

        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
            ValidationError: In case of provided parameters are invalid.
        """
        metadata = get_service_action_metadata(self.create_stream)
        parameters_formatted = to_model_optional(parameters, TextGenerationParameters)
        moderations_formatted = to_model_optional(moderations, ModerationParameters)
        template_formatted = to_model_optional(data, PromptTemplateData)

        self._log_method_execution(
            "Generate Create Stream",
            input=input,
            parameters=parameters_formatted,
            moderations=moderations_formatted,
            template=template_formatted,
        )

        with self._get_http_client() as client:
            yield from generation_stream_handler(
                ResponseModel=TextGenerationStreamCreateResponse,
                logger=self._logger,
                generator=client.post_stream(
                    url=self._get_endpoint(metadata.endpoint),
                    params=_TextGenerationStreamCreateParametersQuery().model_dump(),
                    json=_TextGenerationStreamCreateRequest(
                        input=input,
                        parameters=parameters_formatted,
                        model_id=model_id,
                        prompt_id=prompt_id,
                        moderations=moderations_formatted,
                        data=template_formatted,
                    ).model_dump(),
                ),
            )

    @set_service_action_metadata(endpoint=TextGenerationComparisonCreateEndpoint)
    def compare(
        self,
        *,
        request: TextGenerationComparisonCreateRequestRequest,
        compare_parameters: Optional[ModelLike[TextGenerationComparisonParameters]] = None,
        name: Optional[str] = None,
    ) -> TextGenerationComparisonCreateResponse:
        """
        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
            ValidationError: In case of provided parameters are invalid.
        """
        metadata = get_service_action_metadata(self.compare)
        request_formatted = to_model_instance(request, TextGenerationComparisonCreateRequestRequest)
        compare_parameters_formatted = to_model_instance(compare_parameters, TextGenerationComparisonParameters)

        self._log_method_execution(
            "Text Generation Compare",
            input=input,
            requests=request_formatted,
            parameters=compare_parameters_formatted,
        )

        with self._get_http_client() as client:
            http_response = client.post(
                url=self._get_endpoint(metadata.endpoint),
                params=_TextGenerationComparisonCreateParametersQuery().model_dump(),
                json=_TextGenerationComparisonCreateRequest(
                    name=name,
                    compare_parameters=compare_parameters_formatted,
                    request=request_formatted,
                ).model_dump(),
            )
            return TextGenerationComparisonCreateResponse(**http_response.json())

    def _get_local_limiter(self, limit: Optional[int]):
        return LoopBoundLimiter(lambda: LocalLimiter(limit=limit)) if limit else None



from genai._utils.limiters.adjustable_semaphor import AdjustableAsyncSemaphore
from genai._utils.limiters.base_limiter import BaseLimiter


class LocalLimiter(BaseLimiter):
    """
    A class for limiting the concurrency of certain operations locally with dynamic limit based on report loop feedback
    """

    def __init__(self, limit: int):
        assert limit > 0
        self._max_limit = limit
        self._sp = AdjustableAsyncSemaphore(limit)

    async def acquire(self):
        await self._sp.acquire()
        return True

    def release(self):
        self._sp.release()

    async def report_error(self):
        """Reports an error to the limiter, reducing the current limit by 1 if it is greater than 1."""
        current_limit = self._sp.limit
        if current_limit > 1:
            self._sp.change_max_limit(current_limit - 1)

    async def report_success(self):
        """Reports a successful operation to the limiter, increasing the current limit by 1, up to the maximum limit."""
        new_limit = min(self._max_limit, self._sp.limit + 1)
        self._sp.change_max_limit(new_limit)



from typing import Optional

from genai._utils.service import (
    BaseService,
    BaseServiceConfig,
    BaseServiceServices,
    get_service_action_metadata,
    set_service_action_metadata,
)
from genai.schema import (
    UserCreateEndpoint,
    UserCreateResponse,
    UserDeleteEndpoint,
    UserPatchEndpoint,
    UserPatchResponse,
    UserRetrieveEndpoint,
    UserRetrieveResponse,
)
from genai.schema._api import (
    _UserCreateParametersQuery,
    _UserCreateRequest,
    _UserDeleteParametersQuery,
    _UserPatchParametersQuery,
    _UserPatchRequest,
    _UserRetrieveParametersQuery,
)

__all__ = ["UserService"]


class UserService(BaseService[BaseServiceConfig, BaseServiceServices]):
    @set_service_action_metadata(endpoint=UserCreateEndpoint)
    def create(
        self,
        *,
        first_name: Optional[str] = None,
        last_name: Optional[str] = None,
    ) -> UserCreateResponse:
        """
        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
        """
        request_body = _UserCreateRequest(first_name=first_name, last_name=last_name).model_dump()
        self._log_method_execution("User Create", **request_body)

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.create)
            response = client.post(
                url=self._get_endpoint(metadata.endpoint),
                json=request_body,
                params=_UserCreateParametersQuery().model_dump(),
            )
            return UserCreateResponse(**response.json())

    @set_service_action_metadata(endpoint=UserRetrieveEndpoint)
    def retrieve(self) -> UserRetrieveResponse:
        """
        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
        """
        self._log_method_execution("User Retrieve")

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.retrieve)
            response = client.get(
                url=self._get_endpoint(metadata.endpoint), params=_UserRetrieveParametersQuery().model_dump()
            )
            return UserRetrieveResponse(**response.json())

    @set_service_action_metadata(endpoint=UserPatchEndpoint)
    def update(self, *, tou_accepted: Optional[bool] = None) -> UserPatchResponse:
        """
        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
        """
        request_body = _UserPatchRequest(tou_accepted=tou_accepted).model_dump()
        self._log_method_execution("User Update", **request_body)

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.update)
            response = client.patch(
                url=self._get_endpoint(metadata.endpoint),
                json=request_body,
                params=_UserPatchParametersQuery().model_dump(),
            )
            return UserPatchResponse(**response.json())

    @set_service_action_metadata(endpoint=UserDeleteEndpoint)
    def delete(self) -> None:
        """
        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
        """
        self._log_method_execution("User Delete")

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.delete)
            client.delete(url=self._get_endpoint(metadata.endpoint), params=_UserDeleteParametersQuery().model_dump())



from typing import Any, Optional, Union

from genai._types import ApiBaseModel
from genai._utils.general import to_model_instance
from genai.schema import (
    TextChatCreateResponse,
    TextChatStreamCreateResponse,
    TextGenerationCreateResponse,
    TextGenerationParameters,
    TextGenerationStreamCreateResponse,
)


def extract_token_usage(result: dict[str, Any]):
    def get_count_value(key: str) -> int:
        return result.get(key, 0) or 0

    input_token_count = get_count_value("input_token_count")
    generated_token_count = get_count_value("generated_token_count")

    return {
        "prompt_tokens": input_token_count,
        "completion_tokens": generated_token_count,
        "total_tokens": input_token_count + generated_token_count,
    }


def create_generation_info_from_response(
    response: Union[
        TextGenerationCreateResponse,
        TextGenerationStreamCreateResponse,
        TextChatCreateResponse,
        TextChatStreamCreateResponse,
    ],
    *,
    result: ApiBaseModel,
) -> dict[str, Any]:
    result_meta = result.model_dump(exclude={"generated_text"}, exclude_none=True)

    return create_generation_info(
        meta=response.model_dump(exclude={"results", "moderation"}, exclude_none=True),
        token_usage=extract_token_usage(result_meta),
        **result_meta,
    )


def create_generation_info(
    *, meta: Optional[dict] = None, token_usage: Optional[dict] = None, **kwargs
) -> dict[str, Any]:
    return {
        "meta": meta or {},
        "token_usage": token_usage or extract_token_usage({}),
        **kwargs,
    }


def _prepare_generation_request(
    parameters: Optional[TextGenerationParameters] = None, *, stop: Optional[list[str]] = None, **kwargs: Any
):
    parameters = to_model_instance(parameters, TextGenerationParameters, copy=False)
    if stop is not None:
        parameters.stop_sequences = stop

    request = kwargs.copy()
    request["parameters"] = parameters

    if request.get("prompt_id") is not None:
        request.pop("input", None)
    elif request.get("input") is not None:
        request.pop("prompt_id", None)

    return request


def _prepare_chat_generation_request(**kwargs: Any):
    request = _prepare_generation_request(**kwargs)

    if request.get("conversation_id") is not None:
        request.pop("model_id", None)
        request.pop("prompt_id", None)

    if request.get("prompt_id") is not None:
        request.pop("model_id", None)

    if request.get("use_conversation_parameters") is True:
        request.pop("parameters", None)

    if request.get("parameters") is not None:
        request.pop("use_conversation_parameters", None)

    return request



from typing import Generator, Optional, Union

from genai._types import ModelLike
from genai._utils.async_executor import execute_async
from genai._utils.general import (
    batch_by_size_constraint,
    cast_list,
    merge_objects,
    to_model_instance,
)
from genai._utils.http_client.httpx_client import AsyncHttpxClient
from genai._utils.limiters.local_limiter import LocalLimiter
from genai._utils.service import (
    BaseService,
    BaseServiceConfig,
    BaseServiceServices,
    CommonExecutionOptions,
    get_service_action_metadata,
    set_service_action_metadata,
)
from genai.schema import (
    TextTokenizationCreateEndpoint,
    TextTokenizationCreateResponse,
    TextTokenizationParameters,
)
from genai.schema._api import (
    _TextTokenizationCreateParametersQuery,
    _TextTokenizationCreateRequest,
)

__all__ = ["CreateExecutionOptions", "BaseConfig", "TokenizationService"]

from genai._utils.limiters.shared_limiter import LoopBoundLimiter


class CreateExecutionOptions(BaseServiceConfig):
    """Execution options for tokenization process."""

    throw_on_error: CommonExecutionOptions.throw_on_error = True
    ordered: CommonExecutionOptions.ordered = True
    concurrency_limit: CommonExecutionOptions.concurrency_limit = None
    batch_size: CommonExecutionOptions.batch_size = None
    rate_limit_options: CommonExecutionOptions.rate_limit_options = None
    callback: CommonExecutionOptions.callback[TextTokenizationCreateResponse] = None


class BaseConfig(BaseServiceConfig):
    create_execution_options: CreateExecutionOptions = CreateExecutionOptions()


class TokenizationService(BaseService[BaseConfig, BaseServiceServices]):
    Config = BaseConfig

    @set_service_action_metadata(endpoint=TextTokenizationCreateEndpoint)
    def create(
        self,
        *,
        input: Union[str, list[str]],
        model_id: Optional[str] = None,
        prompt_id: Optional[str] = None,
        parameters: Optional[ModelLike[TextTokenizationParameters]] = None,
        execution_options: Optional[ModelLike[CreateExecutionOptions]] = None,
    ) -> Generator[TextTokenizationCreateResponse, None, None]:
        """
        Args:
            input: The input data for tokenization. It can be a single string or a list of strings.
            model_id: The ID of the model to use for tokenization. Eiter 'model_id' or 'prompt_id' must be provided.
            prompt_id: The ID of the prompt to use for tokenization.  Eiter 'model_id' or 'prompt_id' must be provided.
            parameters: The parameters for tokenization, like return options.
            execution_options: The execution options for tokenization like batch size, callbacks and cetra.

        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
            ValidationError: In case of provided parameters are invalid.
        """
        metadata = get_service_action_metadata(self.create)
        prompts = cast_list(input)
        options = to_model_instance([self.config.create_execution_options, execution_options], CreateExecutionOptions)
        parameters_validated = to_model_instance(parameters, TextTokenizationParameters)
        batches = batch_by_size_constraint(
            prompts,
            max_size_bytes=self._api_client.config.max_payload_size_bytes,
            max_chunk_size=options.batch_size or len(prompts),
        )

        self._log_method_execution(
            "Tokenization Create",
            prompts=prompts,
            parameters=parameters_validated,
            options=options,
        )

        async def handler(inputs_chunk: list[str], http_client: AsyncHttpxClient, *_) -> TextTokenizationCreateResponse:
            http_response = await http_client.post(
                url=self._get_endpoint(metadata.endpoint),
                params=_TextTokenizationCreateParametersQuery().model_dump(),
                json=_TextTokenizationCreateRequest(
                    input=inputs_chunk,
                    model_id=model_id,
                    prompt_id=prompt_id,
                    parameters=parameters_validated,
                ).model_dump(),
            )
            response = TextTokenizationCreateResponse(**http_response.json())

            if options.callback:
                options.callback(response)

            return response

        yield from execute_async(
            inputs=list(batches),
            handler=handler,
            ordered=options.ordered,
            limiters=[self._get_local_limiter(options.concurrency_limit)],
            throw_on_error=options.throw_on_error,
            http_client=lambda: self._get_async_http_client(
                rate_limit_options=merge_objects(
                    {"max_rate": 25, "time_period": 1, "disable_rate_limit_no_header": True},
                    options.rate_limit_options,
                )
            ),
        )

    def _get_local_limiter(self, limit: Optional[int]):
        return LoopBoundLimiter(lambda: LocalLimiter(limit=limit)) if limit else None



from typing import Generator, Optional, Union

import httpx
from httpx import AsyncClient, HTTPStatusError
from pydantic import BaseModel

from genai._types import ModelLike
from genai._utils.api_client import ApiClient
from genai._utils.async_executor import execute_async
from genai._utils.general import cast_list, to_model_instance, to_model_optional
from genai._utils.service import (
    BaseService,
    BaseServiceConfig,
    BaseServiceServices,
    get_service_action_metadata,
    set_service_action_metadata,
)
from genai._utils.shared_options import CommonExecutionOptions
from genai.schema import TextEmbeddingCreateEndpoint, TextEmbeddingCreateResponse, TextEmbeddingParameters
from genai.schema._api import (
    _TextEmbeddingCreateParametersQuery,
    _TextEmbeddingCreateRequest,
)
from genai.text.embedding.limit.limit_service import LimitService as _LimitService

__all__ = ["EmbeddingService", "BaseConfig", "BaseServices", "CreateExecutionOptions"]

from genai._utils.http_client.retry_transport import BaseRetryTransport
from genai._utils.limiters.base_limiter import BaseLimiter
from genai._utils.limiters.external_limiter import ConcurrencyResponse, ExternalLimiter
from genai._utils.limiters.local_limiter import LocalLimiter
from genai._utils.limiters.shared_limiter import LoopBoundLimiter


class BaseServices(BaseServiceServices):
    LimitService: type[_LimitService] = _LimitService


class CreateExecutionOptions(BaseModel):
    throw_on_error: CommonExecutionOptions.throw_on_error = True
    ordered: CommonExecutionOptions.ordered = True
    concurrency_limit: CommonExecutionOptions.concurrency_limit = None
    callback: CommonExecutionOptions.callback[TextEmbeddingCreateResponse] = None


class BaseConfig(BaseServiceConfig):
    create_execution_options: CreateExecutionOptions = CreateExecutionOptions()


class EmbeddingService(BaseService[BaseConfig, BaseServices]):
    Config = BaseConfig
    Services = BaseServices

    def __init__(
        self,
        *,
        api_client: ApiClient,
        services: Optional[BaseServices] = None,
        config: Optional[Union[BaseConfig, dict]] = None,
    ):
        super().__init__(api_client=api_client, config=config)

        if not services:
            services = BaseServices()

        self._concurrency_limiter = self._get_concurrency_limiter()
        self.limit = services.LimitService(api_client=api_client)

    @set_service_action_metadata(endpoint=TextEmbeddingCreateEndpoint)
    def create(
        self,
        *,
        model_id: str,
        inputs: Union[str, list[str]],
        parameters: Optional[ModelLike[TextEmbeddingParameters]] = None,
        execution_options: Optional[ModelLike[CreateExecutionOptions]] = None,
    ) -> Generator[TextEmbeddingCreateResponse, None, None]:
        """
        Creates embedding vectors from an input(s).

        Args:
            model_id: The ID of the model.
            inputs: Text/texts to process. It is recommended not to leave any trailing spaces.
            parameters: Parameters for embedding.
            execution_options: An optional configuration how SDK should work (error handling, limits, callbacks, ...)


        Example::

            from genai import Client, Credentials
            from genai.text.chat import HumanMessage, TextGenerationParameters

            client = Client(credentials=Credentials.from_env())

            responses = list(
                client.text.embedding.create(
                    model_id="sentence-transformers/all-minilm-l6-v2",
                    input="Write a tagline for an alumni association: Together we"
                )
            )
            print("Output vector", responses[0].results[0])

        Yields:
            TextEmbeddingCreateResponse object.

        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
            ValidationError: In case of provided parameters are invalid.
        """
        metadata = get_service_action_metadata(self.create)
        prompts: list[str] = cast_list(inputs)
        parameters_formatted = to_model_optional(parameters, TextEmbeddingParameters)
        execution_options_formatted = to_model_instance(
            [self.config.create_execution_options, execution_options], CreateExecutionOptions
        )
        assert execution_options_formatted

        self._log_method_execution(
            "Embedding Create",
            prompts=prompts,
            execution_options=execution_options_formatted,
        )

        async def handler(input: str, http_client: AsyncClient, limiter: BaseLimiter) -> TextEmbeddingCreateResponse:
            self._log_method_execution("Embedding Create - processing input", input=input)

            async def handle_retry(ex: Exception):
                if isinstance(ex, HTTPStatusError) and ex.response.status_code == httpx.codes.TOO_MANY_REQUESTS:
                    await limiter.report_error()

            async def handle_success(*args):
                await limiter.report_success()

            http_response = await http_client.post(
                url=self._get_endpoint(metadata.endpoint),
                extensions={
                    BaseRetryTransport.Callback.Retry: handle_retry,
                    BaseRetryTransport.Callback.Success: handle_success,
                },
                params=_TextEmbeddingCreateParametersQuery().model_dump(),
                json=_TextEmbeddingCreateRequest(
                    input=input, model_id=model_id, parameters=parameters_formatted
                ).model_dump(),
            )
            response = TextEmbeddingCreateResponse(**http_response.json())

            if execution_options_formatted.callback:
                execution_options_formatted.callback(response)

            return response

        yield from execute_async(
            inputs=prompts,
            handler=handler,
            limiters=[
                self._concurrency_limiter,
                self._get_local_limiter(execution_options_formatted.concurrency_limit),
            ],
            http_client=self._get_async_http_client,
            ordered=execution_options_formatted.ordered,
            throw_on_error=execution_options_formatted.throw_on_error,
        )

    def _get_local_limiter(self, limit: Optional[int]):
        return LoopBoundLimiter(lambda: LocalLimiter(limit=limit)) if limit else None

    def _get_concurrency_limiter(self) -> LoopBoundLimiter:
        async def handler():
            response = await self.limit.aretrieve()
            return ConcurrencyResponse(
                limit=response.result.concurrency.limit,
                remaining=response.result.concurrency.remaining,
            )

        return LoopBoundLimiter(lambda: ExternalLimiter(handler=handler))



import logging

from genai._utils.api_client import ApiClient
from genai._utils.shared_loop import handle_shutdown_event
from genai._version import __version__ as v
from genai.client import Client
from genai.credentials import Credentials

__version__ = v


__all__ = ["Client", "ApiClient", "Credentials", "handle_shutdown_event", "__version__"]



import json
import logging
from typing import Generator, Optional, Type, TypeVar

from pydantic import BaseModel

TModel = TypeVar("TModel", bound=BaseModel)


def generation_stream_handler(
    generator: Generator[Optional[str], None, None], *, logger: logging.Logger, ResponseModel: Type[TModel]
) -> Generator[TModel, None, None]:
    for response in generator:
        if not response:
            continue

        try:
            parsed_response: dict = json.loads(response)
            yield ResponseModel(**parsed_response)
        except Exception as err:
            logger.error("Could not parse {} as json".format(response))
            raise err



from typing import Optional, TypeVar

from pydantic import BaseModel

from genai._types import EnumLike
from genai._utils.general import to_enum
from genai._utils.service import (
    BaseService,
    BaseServiceConfig,
    BaseServiceServices,
    get_service_action_metadata,
    set_service_action_metadata,
)
from genai._utils.validators import assert_is_not_empty_string
from genai.schema import (
    TextGenerationFeedbackCategory,
    TextGenerationIdFeedbackCreateEndpoint,
    TextGenerationIdFeedbackCreateResponse,
    TextGenerationIdFeedbackRetrieveEndpoint,
    TextGenerationIdFeedbackRetrieveResponse,
    TextGenerationIdFeedbackUpdateEndpoint,
    TextGenerationIdFeedbackUpdateResponse,
)
from genai.schema._api import (
    _TextGenerationIdFeedbackCreateParametersQuery,
    _TextGenerationIdFeedbackCreateRequest,
    _TextGenerationIdFeedbackRetrieveParametersQuery,
    _TextGenerationIdFeedbackUpdateParametersQuery,
    _TextGenerationIdFeedbackUpdateRequest,
)

T = TypeVar("T", bound=BaseModel)

__all__ = ["FeedbackService"]


class FeedbackService(BaseService[BaseServiceConfig, BaseServiceServices]):
    @set_service_action_metadata(endpoint=TextGenerationIdFeedbackRetrieveEndpoint)
    def retrieve(
        self,
        generation_id: str,
    ) -> TextGenerationIdFeedbackRetrieveResponse:
        """
        Retrieve feedback for the generated output in the given response.

        Raises:
            ApiResponseException: If target feedback/generation does not exist or cannot be updated.
            ApiNetworkException: In case of unhandled network error.
        """
        metadata = get_service_action_metadata(self.retrieve)
        assert_is_not_empty_string(generation_id)
        self._log_method_execution("Text Generation Feedback Retrieve", generation_id=generation_id)

        with self._get_http_client() as client:
            http_response = client.get(
                url=self._get_endpoint(metadata.endpoint, id=generation_id),
                params=_TextGenerationIdFeedbackRetrieveParametersQuery().model_dump(),
            )
            return TextGenerationIdFeedbackRetrieveResponse(**http_response.json())

    @set_service_action_metadata(endpoint=TextGenerationIdFeedbackCreateEndpoint)
    def create(
        self,
        generation_id: str,
        *,
        categories: Optional[list[EnumLike[TextGenerationFeedbackCategory]]] = None,
        comment: Optional[str] = None,
    ) -> TextGenerationIdFeedbackCreateResponse:
        """
        Provide feedback on generated output for further improvement of the models.

        Args:
            generation_id: A string representing the ID of the text generation.
            categories: An optional list of enum-like objects representing the feedback categories.
            comment: An optional string representing the feedback comment.

        Raises:
            ApiResponseException: If target feedback/generation does not exist or cannot be updated.
            ApiNetworkException: In case of unhandled network error.
        """
        assert_is_not_empty_string(generation_id)
        self._log_method_execution(
            "Text Generation Feedback Create", generation_id=generation_id, categories=categories, comment=comment
        )

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.create)
            http_response = client.post(
                url=self._get_endpoint(metadata.endpoint, id=generation_id),
                params=_TextGenerationIdFeedbackCreateParametersQuery().model_dump(),
                json=_TextGenerationIdFeedbackCreateRequest(
                    comment=comment,
                    categories=(
                        [to_enum(TextGenerationFeedbackCategory, category) for category in categories]
                        if categories is not None
                        else None
                    ),
                ).model_dump(),
            )
            return TextGenerationIdFeedbackCreateResponse(**http_response.json())

    @set_service_action_metadata(endpoint=TextGenerationIdFeedbackUpdateEndpoint)
    def update(
        self,
        generation_id: str,
        categories: Optional[list[TextGenerationFeedbackCategory]] = None,
        comment: Optional[str] = None,
    ) -> TextGenerationIdFeedbackUpdateResponse:
        """
        Update existing feedback.

        Args:
            generation_id: The ID of the text generation to update.
            categories: Optional. List of text generation feedback categories.
            comment: Optional. Comment for the text generation feedback.

        Raises:
            ApiResponseException: If target feedback/generation does not exist or cannot be updated.
            ApiNetworkException: In case of unhandled network error.
        """
        assert_is_not_empty_string(generation_id)
        self._log_method_execution(
            "Text Generation Feedback Update", generation_id=generation_id, categories=categories, comment=comment
        )

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.update)
            http_response = client.put(
                url=self._get_endpoint(metadata.endpoint, id=generation_id),
                params=_TextGenerationIdFeedbackUpdateParametersQuery().model_dump(),
                json=_TextGenerationIdFeedbackUpdateRequest(comment=comment, categories=categories).model_dump(),
            )
            return TextGenerationIdFeedbackUpdateResponse(**http_response.json())



import asyncio
import logging
from asyncio import AbstractEventLoop, Task
from collections.abc import Iterator
from concurrent.futures import CancelledError, Future
from contextlib import suppress
from typing import Awaitable, Callable, Generator, Generic, Optional, TypeVar, Union

from genai._utils.general import first_defined, single_execution
from genai._utils.http_client.httpx_client import AsyncHttpxClient
from genai._utils.limiters.base_limiter import BaseLimiter
from genai._utils.limiters.container_limiter import LimiterContainer
from genai._utils.limiters.shared_limiter import LoopBoundLimiter
from genai._utils.queues.flushable_queue import FlushableQueue
from genai._utils.queues.ordered_queue import OrderedQueue

__all__ = ["execute_async", "BaseConfig"]

from genai._utils.shared_loop import shared_event_loop

TInput = TypeVar("TInput")
TResult = TypeVar("TResult")


logger = logging.getLogger(__name__)


class BaseConfig:
    __slots__ = ()
    ordered: bool = False
    throw_on_error: bool = True
    limit_reach_retry_threshold: float = 1.0
    concurrency_limit: Optional[int] = None


class _AsyncGenerator(Generic[TInput, TResult]):
    """
    Utility iterator to process 'inputs' asynchronously by spawning new thread with it's own event loop.
    Communication is done via Queues.
    """

    def __init__(
        self,
        *,
        inputs: list[TInput],
        http_client: Callable[[], AsyncHttpxClient],
        handler: Callable[[TInput, AsyncHttpxClient, BaseLimiter], Awaitable[TResult]],
        limiters: Optional[list[Optional[LoopBoundLimiter]]] = None,
        ordered: Optional[bool] = None,
        throw_on_error: Optional[bool] = None,
    ):
        self._inputs = inputs
        self._http_client_factory = http_client
        self._handler = handler
        self._limiters = limiters
        self._ordered = first_defined(ordered, default=BaseConfig.ordered)
        self._throw_on_error = first_defined(throw_on_error, default=BaseConfig.throw_on_error)

        self._queue: Union[OrderedQueue, FlushableQueue] = OrderedQueue() if ordered else FlushableQueue()
        self._irrecoverable_error = False
        self._future: Optional[Future] = None

    def _add_to_queue(
        self,
        *,
        idx: Optional[int],
        result: Optional[TResult],
        error: Optional[Exception],
    ):
        entry = idx, result, error
        self._queue.put_nowait(entry)

    async def _process_input(
        self,
        limiter: BaseLimiter,
        batch_num: int,
        input: TInput,
        client: AsyncHttpxClient,
    ):
        async with limiter:
            logger.debug(f"Creating task for batch_num: {batch_num}")
            try:
                response = await self._handler(input, client, limiter)
                logger.debug("Received response = {}".format(response))
                self._add_to_queue(idx=batch_num, result=response, error=None)
            except Exception as e:
                logger.error(f"Exception raised during processing\n{str(e)}")
                self._add_to_queue(idx=batch_num, result=None, error=e)

    async def _schedule_requests(self, limiter: BaseLimiter, loop: AbstractEventLoop):
        tasks: list[Task] = []
        try:
            async with self._http_client_factory() as client:
                tasks.extend(
                    loop.create_task(self._process_input(limiter, idx, input, client))
                    for idx, input in enumerate(self._inputs)
                )
                await asyncio.gather(*tasks)
        except Exception as ex:
            self._irrecoverable_error = True
            self._add_to_queue(idx=None, result=None, error=ex)
            raise ex
        finally:
            for task in tasks:
                task.cancel()

    @single_execution
    def _handle_close_signal(self):
        if not self._future:
            return
        self._future.cancel()
        self._queue.flush()
        self._irrecoverable_error = True
        self._add_to_queue(
            idx=None,
            result=None,
            error=InterruptedError("Generation has been aborted by the user."),
        )

    def create_iterator(self) -> Iterator[TResult]:
        if not self._inputs:
            return
        with shared_event_loop as loop:
            limiter = LimiterContainer(*(self._limiters or []))
            self._future = asyncio.run_coroutine_threadsafe(self._schedule_requests(limiter, loop), loop)

            shared_event_loop.add_close_handler(self._handle_close_signal)
            try:
                for _ in enumerate(self._inputs):
                    batch_num, response, error = self._queue.get()
                    self._queue.task_done()

                    if (self._throw_on_error or self._irrecoverable_error) and error:
                        raise error

                    yield response
            except Exception:
                # future is PENDING even if there are running tasks due to implementation of run_coroutine_threadsafe
                # therefore cancel always succeeds
                self._future.cancel()
                raise
            finally:
                with suppress(CancelledError):
                    self._future.result()
                self._queue.flush()
                shared_event_loop.remove_close_handler(self._handle_close_signal)

    def __iter__(self) -> Iterator[TResult]:
        return self.create_iterator()


def execute_async(
    *,
    inputs: list[TInput],
    http_client: Callable[[], AsyncHttpxClient],
    handler: Callable[[TInput, AsyncHttpxClient, BaseLimiter], Awaitable[TResult]],
    limiters: Optional[list[Optional[LoopBoundLimiter]]] = None,
    ordered: Optional[bool] = None,
    throw_on_error: Optional[bool] = None,
) -> Generator[TResult, None, None]:
    yield from _AsyncGenerator(
        inputs=inputs,
        http_client=http_client,
        handler=handler,
        limiters=limiters,
        ordered=ordered,
        throw_on_error=throw_on_error,
    )



from typing import Callable, TypeVar

from pydantic import BaseModel, ConfigDict, field_serializer

from genai.schema._endpoints import ApiEndpoint

__all__ = ["set_service_action_metadata", "get_service_action_metadata", "inherit_metadata"]


class ServiceActionMetadata(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
        validate_default=True,
        validate_assignment=True,
        validate_return=True,
    )

    endpoint: type[ApiEndpoint]

    @field_serializer("endpoint")
    def serialize_endpoint(self, dt: type[ApiEndpoint], _info):
        return {k: v for k, v in vars(dt).items() if not k.startswith("__")}


_METADATA_KEY = "_metadata"

T = TypeVar("T", bound=Callable)


def inherit_metadata(*, source: Callable, target: Callable) -> None:
    source_meta = getattr(source, _METADATA_KEY, None)
    target_meta = getattr(target, _METADATA_KEY, None)

    if source_meta is not None and target_meta is None:
        setattr(target, _METADATA_KEY, source_meta)


def set_service_action_metadata(*, endpoint: type[ApiEndpoint]):
    def decorator(fn: T) -> T:
        setattr(fn, _METADATA_KEY, ServiceActionMetadata(endpoint=endpoint))
        return fn

    return decorator


def get_service_action_metadata(service_method: Callable) -> ServiceActionMetadata:
    """
    Retrieve metadata for arbitrary service's method.
    For instance which endpoint is being used.

    Example::

        from genai import Credentials, Client
        from genai.utils import get_service_action_metadata

        credentials = Credentials.from_env()
        client = Client(credentials=credentials)

        metadata = get_service_action_metadata(client.text.tokenization.create)
    """

    metadata = getattr(service_method, _METADATA_KEY, None)
    if not isinstance(metadata, ServiceActionMetadata):
        raise ValueError("Provided callable does not have any metadata.")

    return metadata



from typing import Any, Optional

import httpx
from typing_extensions import TypeGuard

from genai.schema import (
    BadRequestResponse,
    BaseErrorResponse,
    InternalServerErrorResponse,
    NotFoundResponse,
    TooManyRequestsResponse,
    UnauthorizedResponse,
)

__all__ = [
    "is_api_error_response",
    "get_api_error_class_by_status_code",
    "to_api_error",
    "BaseErrorResponse",
]


def is_api_error_response(input: Any) -> TypeGuard[BaseErrorResponse]:
    return isinstance(input, BaseErrorResponse)


def get_api_error_class_by_status_code(code: int) -> Optional[type[BaseErrorResponse]]:
    response_class_mapping: dict[int, type[BaseErrorResponse]] = {
        httpx.codes.TOO_MANY_REQUESTS.value: TooManyRequestsResponse,
        httpx.codes.NOT_FOUND.value: NotFoundResponse,
        httpx.codes.UNAUTHORIZED.value: UnauthorizedResponse,
        httpx.codes.INTERNAL_SERVER_ERROR.value: InternalServerErrorResponse,
        httpx.codes.BAD_REQUEST.value: BadRequestResponse,
    }
    return response_class_mapping.get(code)


def to_api_error(body: dict) -> BaseErrorResponse:
    cls: type[BaseErrorResponse] = get_api_error_class_by_status_code(body["status_code"]) or BaseErrorResponse
    model = cls.model_validate(body)
    return model



from genai.schema._api import BaseMessage, ChatRole


class HumanMessage(BaseMessage):
    role: ChatRole = ChatRole.USER


class SystemMessage(BaseMessage):
    role: ChatRole = ChatRole.SYSTEM


class AIMessage(BaseMessage):
    role: ChatRole = ChatRole.ASSISTANT


__all__ = ["HumanMessage", "SystemMessage", "AIMessage"]



"""Extension for LangChain library"""
import warnings
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as get_package_version

from packaging.version import parse as parse_version

from genai.extensions.langchain.chat_llm import LangChainChatInterface
from genai.extensions.langchain.embeddings import LangChainEmbeddingsInterface
from genai.extensions.langchain.llm import LangChainInterface
from genai.extensions.langchain.template import (
    from_langchain_template,
    to_langchain_template,
)


def _verify_langchain_version():
    try:
        langchain_version = get_package_version("langchain")
        minimal_valid_version = parse_version("0.1.0")
        current_version = parse_version(langchain_version)

        if current_version < minimal_valid_version:
            warnings.warn(
                category=RuntimeWarning,
                stacklevel=4,
                message=f"You are using 'langchain' package version {current_version}. "
                f"Please do upgrade now to version >= {minimal_valid_version}.",
            )

    except PackageNotFoundError:
        pass


_verify_langchain_version()

__all__ = [
    "LangChainInterface",
    "LangChainChatInterface",
    "LangChainEmbeddingsInterface",
    "from_langchain_template",
    "to_langchain_template",
]



"""Wrapper around IBM GENAI APIs for use in Langchain"""

import asyncio
import logging
from functools import partial
from pathlib import Path
from typing import Any, Iterator, Optional, Union

from pydantic import ConfigDict
from pydantic.v1 import validator

from genai import Client
from genai._utils.general import to_model_instance, to_model_optional
from genai.extensions._common.utils import (
    _prepare_generation_request,
    create_generation_info,
    create_generation_info_from_response,
)
from genai.schema import (
    ModerationParameters,
    PromptTemplateData,
    TextGenerationParameters,
    TextGenerationStreamCreateResponse,
)
from genai.text.generation import CreateExecutionOptions

try:
    from langchain_core.callbacks.manager import (
        AsyncCallbackManagerForLLMRun,
        CallbackManagerForLLMRun,
    )
    from langchain_core.language_models.llms import LLM
    from langchain_core.messages import BaseMessage, get_buffer_string
    from langchain_core.outputs import LLMResult

    from genai.extensions.langchain.utils import (
        CustomGenerationChunk,
        create_llm_output,
        dump_optional_model,
        load_config,
        update_token_usage,
        update_token_usage_stream,
    )
except ImportError:
    raise ImportError("Could not import langchain: Please install ibm-generative-ai[langchain] extension.")  # noqa: B904

logger = logging.getLogger(__name__)

__all__ = ["LangChainInterface"]


class LangChainInterface(LLM):
    """
    Class representing the LangChainChatInterface for interacting with the LangChain chat API.

    Example::

        from genai import Client, Credentials
        from genai.extensions.langchain import LangChainInterface
        from genai.schema import TextGenerationParameters

        client = Client(credentials=Credentials.from_env())
        llm = LangChainInterface(
            client=client,
            model_id="meta-llama/llama-2-70b-chat",
            parameters=TextGenerationParameters(max_new_tokens=50)
        )

        response = chat_model.generate(prompts=["Hello world!"])
        print(response)
    """

    model_config = ConfigDict(extra="forbid", protected_namespaces=())

    client: Client
    model_id: str
    prompt_id: Optional[str] = None
    parameters: Optional[TextGenerationParameters] = None
    moderations: Optional[ModerationParameters] = None
    data: Optional[PromptTemplateData] = None
    streaming: Optional[bool] = None
    execution_options: Optional[CreateExecutionOptions] = None

    @validator("parameters", "moderations", "data", "execution_options", pre=True, always=True)
    @classmethod
    def _validate_data_models(cls, value, values, config, field):
        return to_model_optional(value, Model=field.type_, copy=False)

    @property
    def _common_identifying_params(self):
        return {
            "model_id": self.model_id,
            "prompt_id": self.prompt_id,
            "parameters": dump_optional_model(self.parameters),
            "moderations": dump_optional_model(self.moderations),
            "data": dump_optional_model(self.data),
            **super()._identifying_params,
        }

    @property
    def _identifying_stream_params(self) -> dict[str, Any]:
        return self._common_identifying_params

    @property
    def _identifying_params(self) -> dict[str, Any]:
        return {
            **self._common_identifying_params,
            "execution_options": dump_optional_model(self.execution_options),
        }

    @classmethod
    def is_lc_serializable(cls) -> bool:
        return True

    @property
    def lc_secrets(self) -> dict[str, str]:
        return {"client": "CLIENT"}

    @classmethod
    def load_from_file(cls, file: Union[str, Path], *, client: Client):
        config = load_config(file)
        return cls(**config, client=client)

    @property
    def _llm_type(self) -> str:
        """Return type of llm."""
        return "ibmgenai_llm"

    def _prepare_request(self, **kwargs):
        updated = {k: kwargs.pop(k, v) for k, v in self._identifying_params.items()}
        return _prepare_generation_request(**kwargs, **updated)

    def _prepare_stream_request(self, **kwargs):
        updated = {k: kwargs.pop(k, v) for k, v in self._identifying_stream_params.items()}
        return _prepare_generation_request(**kwargs, **updated)

    def _call(
        self,
        prompt: str,
        stop: Optional[list[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> str:
        result = self._generate(prompts=[prompt], stop=stop, run_manager=run_manager, **kwargs)
        return result.generations[0][0].text

    def _generate(
        self,
        prompts: list[str],
        stop: Optional[list[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> LLMResult:
        final_result = LLMResult(generations=[], llm_output=create_llm_output(model=self.model_id))
        assert final_result.llm_output

        if len(prompts) == 0:
            return final_result

        if self.streaming:
            if len(prompts) != 1:
                raise ValueError("Streaming works only for a single prompt.")

            generation = CustomGenerationChunk(text="", generation_info=create_generation_info())

            for chunk_result in self._stream(
                prompt=prompts[0],
                stop=stop,
                run_manager=run_manager,
                **kwargs,
            ):
                assert chunk_result.generation_info
                token_usage = chunk_result.generation_info.pop("token_usage")
                generation += chunk_result
                assert generation.generation_info
                update_token_usage_stream(
                    target=generation.generation_info["token_usage"],
                    source=token_usage,
                )

            final_result.generations.append([generation])
            assert generation.generation_info
            update_token_usage(
                target=final_result.llm_output["token_usage"], source=generation.generation_info["token_usage"]
            )

            return final_result
        else:
            responses = list(
                self.client.text.generation.create(**self._prepare_request(inputs=prompts, stop=stop, **kwargs))
            )
            for response in responses:
                for result in response.results:
                    generation_info = create_generation_info_from_response(response, result=result)

                    chunk = CustomGenerationChunk(
                        text=result.generated_text or "",
                        generation_info=generation_info,
                    )
                    logger.info("Output of GENAI call: {}".format(chunk.text))
                    update_token_usage(
                        target=final_result.llm_output["token_usage"], source=generation_info["token_usage"]
                    )
                    final_result.generations.append([chunk])

            return final_result

    async def _agenerate(
        self,
        prompts: list[str],
        stop: Optional[list[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> LLMResult:
        return await asyncio.get_running_loop().run_in_executor(
            None, partial(self._generate, prompts, stop, run_manager, **kwargs)
        )

    def _stream(
        self,
        prompt: str,
        stop: Optional[list[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> Iterator[CustomGenerationChunk]:
        params = to_model_instance(self.parameters, TextGenerationParameters)
        params.stop_sequences = stop or params.stop_sequences

        def send_chunk(
            *, text: Optional[str] = None, generation_info: dict, response: TextGenerationStreamCreateResponse
        ):
            logger.info("Chunk received: {}".format(text))
            chunk = CustomGenerationChunk(
                text=text or "",
                generation_info=generation_info.copy(),
            )
            yield chunk
            if run_manager:
                run_manager.on_llm_new_token(token=chunk.text, chunk=chunk, response=response)

        for response in self.client.text.generation.create_stream(
            **self._prepare_stream_request(input=prompt, stop=stop, **kwargs)
        ):
            if response.moderation:
                generation_info = create_generation_info_from_response(response, result=response.moderation)
                yield from send_chunk(generation_info=generation_info, response=response)

            for result in response.results or []:
                generation_info = create_generation_info_from_response(response, result=result)
                yield from send_chunk(text=result.generated_text, generation_info=generation_info, response=response)

    def get_num_tokens(self, text: str) -> int:
        response = list(self.client.text.tokenization.create(model_id=self.model_id, input=[text]))[0]
        return response.results[0].token_count

    def get_num_tokens_from_messages(self, messages: list[BaseMessage]) -> int:
        return sum(
            sum(result.token_count for result in response.results or [] if result.token_count)
            for response in self.client.text.tokenization.create(
                model_id=self.model_id, input=[get_buffer_string([message]) for message in messages]
            )
        )

    def get_token_ids(self, text: str) -> list[int]:
        raise NotImplementedError("API does not support returning token ids.")



from typing import Optional

from genai._types import EnumLike, ModelLike
from genai._utils.general import to_enum, to_enum_optional, to_model_optional
from genai._utils.service import (
    BaseService,
    BaseServiceConfig,
    BaseServiceServices,
    get_service_action_metadata,
    set_service_action_metadata,
)
from genai._utils.validators import assert_is_not_empty_string
from genai.schema import (
    TuneAssetType,
    TuneCreateEndpoint,
    TuneCreateResponse,
    TuneIdContentTypeRetrieveEndpoint,
    TuneIdDeleteEndpoint,
    TuneIdRetrieveEndpoint,
    TuneIdRetrieveResponse,
    TuneParameters,
    TuneRetrieveEndpoint,
    TuneRetrieveResponse,
    TuneStatus,
    TuningType,
    TuningTypeRetrieveEndpoint,
    TuningTypeRetrieveResponse,
)
from genai.schema._api import (
    _TuneCreateParametersQuery,
    _TuneCreateRequest,
    _TuneIdContentTypeRetrieveParametersQuery,
    _TuneIdDeleteParametersQuery,
    _TuneIdRetrieveParametersQuery,
    _TuneRetrieveParametersQuery,
    _TuningTypeRetrieveParametersQuery,
)

__all__ = ["TuneService"]


class TuneService(BaseService[BaseServiceConfig, BaseServiceServices]):
    @set_service_action_metadata(endpoint=TuneCreateEndpoint)
    def create(
        self,
        *,
        model_id: str,
        name: str,
        task_id: str,
        training_file_ids: list[str],
        tuning_type: EnumLike[TuningType],
        validation_file_ids: Optional[list[str]] = None,
        parameters: Optional[ModelLike[TuneParameters]] = None,
    ) -> TuneCreateResponse:
        """
        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
            ValidationError: In case of provided parameters are invalid.
        """
        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.create)
            request_body = _TuneCreateRequest(
                model_id=model_id,
                name=name,
                parameters=to_model_optional(parameters, TuneParameters),
                task_id=task_id,
                training_file_ids=training_file_ids,
                tuning_type=to_enum(TuningType, tuning_type),
                validation_file_ids=validation_file_ids,
            ).model_dump()

            self._log_method_execution("Tune Create", **request_body)
            response = client.post(
                url=self._get_endpoint(metadata.endpoint),
                params=_TuneCreateParametersQuery().model_dump(),
                json=request_body,
            )
            return TuneCreateResponse(**response.json())

    @set_service_action_metadata(endpoint=TuneIdContentTypeRetrieveEndpoint)
    def read(self, *, id: str, type: EnumLike[TuneAssetType]) -> bytes:
        """
        Download tune assets.

        Raises:
            ValueError: if the tune status is not 'COMPLETED'.
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
            ValidationError: In case of provided parameters are invalid.
        """
        assert_is_not_empty_string(id)
        self._log_method_execution("Tune Read", id=id)

        tune = self.retrieve(id).result
        if not tune.status or tune.status != TuneStatus.COMPLETED:
            raise ValueError(
                f"Tune status: '{tune.status.value if tune.status else 'unknown'}'."
                f"Tune can not be downloaded if status is not '{TuneStatus.COMPLETED.value}'."
            )

        metadata = get_service_action_metadata(self.read)
        with self._get_http_client() as client:
            response = client.get(
                url=self._get_endpoint(metadata.endpoint, id=id, type=to_enum(TuneAssetType, type)),
                params=_TuneIdContentTypeRetrieveParametersQuery().model_dump(),
            )
            return response.content

    @set_service_action_metadata(endpoint=TuneIdRetrieveEndpoint)
    def retrieve(
        self,
        id: str,
    ) -> TuneIdRetrieveResponse:
        """
        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
        """
        metadata = get_service_action_metadata(self.retrieve)
        assert_is_not_empty_string(id)
        self._log_method_execution("Tune Retrieve", id=id)

        with self._get_http_client() as client:
            response = client.get(
                url=self._get_endpoint(metadata.endpoint, id=id),
                params=_TuneIdRetrieveParametersQuery().model_dump(),
            )
            return TuneIdRetrieveResponse(**response.json())

    @set_service_action_metadata(endpoint=TuneRetrieveEndpoint)
    def list(
        self,
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        status: Optional[TuneStatus] = None,
        search: Optional[str] = None,
    ) -> TuneRetrieveResponse:
        """
        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
        """
        self._log_method_execution("Tune List")

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.list)
            response = client.get(
                url=self._get_endpoint(metadata.endpoint),
                params=_TuneRetrieveParametersQuery(
                    limit=limit, offset=offset, status=to_enum_optional(status, TuneStatus), search=search
                ).model_dump(),
            )
            return TuneRetrieveResponse(**response.json())

    @set_service_action_metadata(endpoint=TuningTypeRetrieveEndpoint)
    def types(self) -> TuningTypeRetrieveResponse:
        """
        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
        """
        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.types)
            response = client.get(
                url=self._get_endpoint(metadata.endpoint),
                params=_TuningTypeRetrieveParametersQuery().model_dump(),
            )

            self._log_method_execution("Tune Types")
            return TuningTypeRetrieveResponse(**response.json())

    @set_service_action_metadata(endpoint=TuneIdDeleteEndpoint)
    def delete(
        self,
        id: str,
    ) -> None:
        """
        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
        """
        assert_is_not_empty_string(id)
        self._log_method_execution("Tune Delete")

        with self._get_http_client() as client:
            metadata = get_service_action_metadata(self.delete)
            client.delete(
                url=self._get_endpoint(metadata.endpoint, id=id), params=_TuneIdDeleteParametersQuery().model_dump()
            )



from enum import Enum
from typing import Any, TypeVar, Union

from pydantic import BaseModel, ConfigDict
from typing_extensions import Literal


class ApiBaseModel(BaseModel):
    model_config = ConfigDict(
        extra="allow",
        use_enum_values=True,
        protected_namespaces=(),
        validate_assignment=True,
        allow_inf_nan=False,
        validate_default=True,
    )

    def model_dump(
        self,
        *,
        mode: Union[Literal["json", "python"], str] = "python",
        include: Any = None,
        exclude: Any = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        # exclude_none -> change from False to True
        exclude_none: bool = True,
        round_trip: bool = False,
        warnings: bool = True,
    ) -> dict[str, Any]:
        return super().model_dump(
            mode=mode,
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            warnings=warnings,
        )


E = TypeVar("E", bound=Enum)
EnumLike = Union[str, E]
EnumLikeList = list[Union[str, E]]
EnumLikeOrEnumLikeList = Union[EnumLike[E], EnumLikeList[E]]

D = TypeVar("D", bound=BaseModel)
ModelLike = Union[dict, D]



from typing import Generator, Optional, Union

from genai._types import ModelLike
from genai._utils.async_executor import execute_async
from genai._utils.general import cast_list, to_model_instance, to_model_optional
from genai._utils.http_client.httpx_client import AsyncHttpxClient
from genai._utils.service import (
    BaseService,
    BaseServiceConfig,
    BaseServiceServices,
    CommonExecutionOptions,
    get_service_action_metadata,
    set_service_action_metadata,
)
from genai.schema import (
    HAPOptions,
    ImplicitHateOptions,
    StigmaOptions,
    TextModerationCreateEndpoint,
    TextModerationCreateResponse,
)
from genai.schema._api import (
    _TextModerationCreateParametersQuery,
    _TextModerationCreateRequest,
)

__all__ = ["ModerationService", "BaseConfig", "CreateExecutionOptions"]


class CreateExecutionOptions(BaseServiceConfig):
    throw_on_error: CommonExecutionOptions.throw_on_error = True
    ordered: CommonExecutionOptions.ordered = True


class BaseConfig(BaseServiceConfig):
    create_execution_options: CreateExecutionOptions = CreateExecutionOptions()


class ModerationService(BaseService[BaseConfig, BaseServiceServices]):
    Config = BaseConfig

    @set_service_action_metadata(endpoint=TextModerationCreateEndpoint)
    def create(
        self,
        inputs: Union[str, list[str]],
        *,
        hap: Optional[ModelLike[HAPOptions]] = None,
        implicit_hate: Optional[ModelLike[ImplicitHateOptions]] = None,
        stigma: Optional[ModelLike[StigmaOptions]] = None,
        execution_options: Optional[ModelLike[CreateExecutionOptions]] = None,
    ) -> Generator[TextModerationCreateResponse, None, None]:
        """
        Args:
            inputs: Prompt/Prompts for text moderation.
            hap: HAP configuration (hate, abuse, profanity).
            implicit_hate: Implicit Hate configuration.
            stigma: Stigma configuration.
            execution_options: Configuration processing.

        Example:
            from genai import Client, Credentials
            from genai.text.moderation import HAPOptions

            client = Client(credentials=Credentials.from_env())
            inputs = ["First input", "Second input"]
            counter = 0

            for response in client.text.moderation.create(inputs=inputs, hap=HAPOptions(threshold=0.65)):
                for result in response.results:
                    input = inputs[counter]
                    counter += 1
                    print(f"Response for {input}", result.hap)

        Raises:
            ApiResponseException: In case of a known API error.
            ApiNetworkException: In case of unhandled network error.
            ValidationError: In case of provided parameters are invalid.
        """
        metadata = get_service_action_metadata(self.create)
        execution_options_formatted = to_model_instance(
            [self.config.create_execution_options, execution_options], CreateExecutionOptions
        )
        self._log_method_execution(
            "Moderation Create",
            prompts=inputs,
            hap=hap,
            implicit_hate=implicit_hate,
            stigma=stigma,
            execution_options=execution_options_formatted,
        )

        async def handler(input: str, http_client: AsyncHttpxClient, *_) -> TextModerationCreateResponse:
            self._log_method_execution("Moderation Create - processing input", input=input)

            http_response = await http_client.post(
                url=self._get_endpoint(metadata.endpoint),
                params=_TextModerationCreateParametersQuery().model_dump(),
                json=_TextModerationCreateRequest(
                    input=input,
                    hap=to_model_optional(hap, HAPOptions),
                    implicit_hate=to_model_optional(hap, ImplicitHateOptions),
                    stigma=to_model_optional(stigma, StigmaOptions),
                ).model_dump(),
            )
            return TextModerationCreateResponse(**http_response.json())

        yield from execute_async(
            inputs=cast_list(inputs),
            handler=handler,
            http_client=self._get_async_http_client,
            ordered=execution_options_formatted.ordered,
            throw_on_error=execution_options_formatted.throw_on_error,
        )
